#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest
from rest_framework.test import APIClient

from dj_rql.drf.backend import FilterCache


@pytest.fixture
def api_client():
    client = APIClient()
    client.default_format = 'json'
    return client


@pytest.fixture
def clear_cache():
    FilterCache.clear()



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from copy import copy

from rest_framework.schemas.openapi import SchemaGenerator

from dj_rql.openapi import RQLFilterClassSpecification, RQLFilterDescriptionTemplate
from tests.dj_rf.filters import BooksFilterClass
from tests.dj_rf.models import Book


def filter_instance():
    return BooksFilterClass(None)


def filter_item(f_name, f_item):
    class Cls(RQLFilterClassSpecification):
        @classmethod
        def enrich_with_openapi_data(cls):
            item = copy(f_item)
            item['oa'] = cls._get_filter_item_openapi_data(f_name, f_item)
            return item

    return Cls.enrich_with_openapi_data()


def filter_data(f_name):
    f_instance = filter_instance()
    f_item = filter_item(f_name, f_instance.filters[f_name])
    return f_item, f_instance


def test_description_common_render():
    result = RQLFilterDescriptionTemplate.render(*filter_data('id'))
    assert result == '**Filter for: id**\n\nlookups: eq, ne, ge, gt, le, lt, null, in, out'


def test_description_search_render():
    result = RQLFilterDescriptionTemplate.render(*filter_data('str_choice_field'))
    assert result == '**Filter for: str_choice_field**\n\nlookups: ' \
                     'eq, ne, like, ilike, in, out\nsearch: true'


def test_description_ordering_render():
    result = RQLFilterDescriptionTemplate.render(*filter_data('int_choice_field'))
    assert result == '**Filter for: int_choice_field**\n\nlookups: ' \
                     'eq, ne, ge, gt, le, lt, in, out\nordering: true'


def test_description_null_overridden_render():
    result = RQLFilterDescriptionTemplate.render(*filter_data('title'))
    assert result == '**Filter for: title**\n\n' \
                     'lookups: eq, ne, like, ilike, null, in, out\nsearch: true\n' \
                     'null: NULL_ID, null()'


def test_description_hidden_render():
    result = RQLFilterDescriptionTemplate.render(*filter_data('select_author'))
    assert result == '**Filter for: select_author**\n\n' \
                     'lookups: eq, ne, like, ilike, in, out\ndefault: **hidden**'


def test_description_custom_render():
    class Cls(RQLFilterDescriptionTemplate):
        IN_PLACE_RENDERERS = ('_render_custom_inplace',)

        @classmethod
        def _render_custom_inplace(cls, *args):
            return 'COMMON'

    result = Cls.render(*filter_data('id'))
    assert result == 'COMMON'


def test_specification_render():
    class Cls(RQLFilterClassSpecification):
        @classmethod
        def get_for_field(cls, filter_item, filter_instance):
            if filter_item['oa']['name'] == 'd_id':
                return {'name': 'd_id', 'custom': True}

    f_instance = filter_instance()
    result = Cls.get(f_instance)

    assert result[0]['name'] == 'amazon_rating'
    assert result[-2]['name'] == 'written'
    assert result[-1]['name'] == 'author__email'

    converted_result = {f['name']: f for f in result}

    assert not converted_result['id']['required']
    assert not converted_result['id']['deprecated']
    assert converted_result['id']['in'] == 'query'
    assert converted_result['id']['schema'] == {'type': 'integer'}

    assert converted_result['status']['required']
    assert converted_result['status']['schema'] == {
        'type': 'string',
        'enum': list(Book.STATUS_CHOICES),
    }

    assert 'Author Email' in converted_result['author__email']['description']
    assert converted_result['author__email']['deprecated']

    assert 'name' not in converted_result
    assert 'author' not in converted_result

    assert converted_result['published.at']['schema'] == {'type': 'string', 'format': 'date-time'}

    assert converted_result['rating.blog']['schema'] == {
        'type': 'string',
        'enum': ['low', 'high'],
    }
    assert converted_result['rating.blog_int']['schema'] == {
        'type': 'integer',
        'enum': [Book.LOW_RATING, Book.HIGH_RATING],
    }

    assert converted_result['d_id'] == {'name': 'd_id', 'custom': True}

    assert converted_result['str_choice_field']['schema'] == {
        'type': 'string',
        'enum': ['one', 'two'],
    }
    assert converted_result['str_choice_field_repr']['schema'] == {
        'type': 'string',
        'enum': ['I', 'II'],
    }

    assert converted_result['anno_int']['schema'] == {'type': 'integer'}
    assert converted_result['author.is_male']['schema'] == {'type': 'boolean'}
    assert converted_result['written']['schema'] == {'type': 'string', 'format': 'date'}
    assert converted_result['current_price']['schema'] == {'type': 'number', 'format': 'double'}
    assert converted_result['amazon_rating']['schema'] == {'type': 'number', 'format': 'float'}

    assert converted_result['url']['schema'] == {'type': 'string'}
    assert converted_result['t__in']['schema'] == {'type': 'custom', 'format': 'custom'}


def test_default_get_for_field():
    assert RQLFilterClassSpecification.get_for_field(None, None) is None


def test_api_auto_genereation():
    openapi_schema = SchemaGenerator().get_schema()

    assert '/books/' in openapi_schema['paths']
    parameters = openapi_schema['paths']['/books/']['get']['parameters']
    assert len(parameters) > 10
    assert parameters[0]['name'] == 'limit'
    assert parameters[0]['schema'] == {'type': 'integer'}
    assert parameters[1]['name'] == 'offset'
    assert parameters[1]['schema'] == {'type': 'integer'}
    assert parameters[2]['name'] == 'amazon_rating'

    assert '/nofiltercls/' in openapi_schema['paths']

    assert '/books/{id}/' not in openapi_schema['paths']

    assert '/books/{id}/act/' in openapi_schema['paths']
    parameters = openapi_schema['paths']['/books/{id}/act/']['get']['parameters']
    assert len(parameters) == 1
    assert parameters[0]['name'] == 'id'

    assert '/select/{id}/' in openapi_schema['paths']
    parameters = openapi_schema['paths']['/select/{id}/']['get']['parameters']
    assert len(parameters) == 1
    assert parameters[0]['name'] == 'id'

    assert '/old_books/{id}/' in openapi_schema['paths']
    get_parameters = openapi_schema['paths']['/old_books/{id}/']['get']['parameters']
    assert len(get_parameters) > 10
    put_parameters = openapi_schema['paths']['/old_books/{id}/']['put']['parameters']
    assert len(put_parameters) == 1
    assert put_parameters[0]['name'] == 'id'



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from datetime import date, datetime
from functools import partial

import pytest
from django.db.models import Q

from dj_rql._dataclasses import FilterArgs
from dj_rql.constants import (
    ComparisonOperators as CO,
    DjangoLookups,
    SearchOperators,
    RQL_EMPTY,
    RQL_NULL,
)
from dj_rql.exceptions import RQLFilterLookupError, RQLFilterParsingError, RQLFilterValueError
from tests.dj_rf.filters import BooksFilterClass
from tests.dj_rf.models import Author, Book, Page, Publisher
from tests.test_filter_cls.utils import book_qs, create_books


def filter_field(filter_name, operator, value):
    filter_cls = BooksFilterClass(book_qs)
    q = filter_cls.build_q_for_filter(FilterArgs(filter_name, operator, str(value)))
    return list(book_qs.filter(q))


def assert_filter_field_error(error_cls, filter_name, operator, value):
    with pytest.raises(error_cls) as e:
        filter_field(filter_name, operator, value)
    assert e.value.details == {
        'filter': filter_name,
        'lookup': operator,
        'value': str(value),
    }


assert_filter_field_value_error = partial(assert_filter_field_error, RQLFilterValueError)
assert_filter_field_lookup_error = partial(assert_filter_field_error, RQLFilterLookupError)


@pytest.mark.django_db
def test_id():
    filter_name = 'id'
    books = create_books()
    assert filter_field(filter_name, CO.EQ, books[0].pk) == [books[0]]
    assert filter_field(filter_name, CO.EQ, 3) == []
    assert filter_field(filter_name, CO.NE, books[1].pk) == [books[0]]
    assert filter_field(filter_name, CO.LT, books[1].pk) == [books[0]]
    assert filter_field(filter_name, CO.GE, books[0].pk) == books


@pytest.mark.django_db
def test_title():
    filter_name = 'title'
    books = [
        Book.objects.create(title='G'),
        Book.objects.create(title='R'),
        Book.objects.create(),
        Book.objects.create(title=''),
    ]
    assert filter_field(filter_name, CO.EQ, books[0].title) == [books[0]]
    assert filter_field(filter_name, CO.EQ, '"{}"'.format(books[0].title)) == [books[0]]
    assert filter_field(filter_name, CO.EQ, "'{}'".format(books[0].title)) == [books[0]]
    assert filter_field(filter_name, CO.EQ, 'N') == []
    assert filter_field(filter_name, CO.NE, books[0].title) == [books[1], books[2], books[3]]
    assert filter_field(filter_name, CO.EQ, RQL_NULL) == [books[2]]
    assert filter_field(filter_name, CO.EQ, 'NULL_ID') == [books[2]]
    assert filter_field(filter_name, CO.NE, 'NULL_ID') == [books[0], books[1], books[3]]
    assert filter_field(filter_name, CO.EQ, RQL_EMPTY) == [books[3]]
    assert filter_field(filter_name, CO.NE, RQL_EMPTY) == [books[0], books[1], books[2]]


@pytest.mark.django_db
def test_current_price():
    filter_name = 'current_price'
    books = [
        Book.objects.create(current_price=5.23),
        Book.objects.create(current_price=0.0121),
    ]
    assert filter_field(filter_name, CO.EQ, books[0].current_price) == [books[0]]
    assert filter_field(filter_name, CO.EQ, 5.2300123) == [books[0]]
    assert filter_field(filter_name, CO.EQ, 2) == []
    assert filter_field(filter_name, CO.NE, books[1].current_price) == [books[0]]
    assert filter_field(filter_name, CO.LE, books[0].current_price) == books
    assert filter_field(filter_name, CO.GT, books[1].current_price) == [books[0]]


@pytest.mark.django_db
def test_written():
    filter_name = 'written'
    books = [
        Book.objects.create(written=date(2019, 2, 12)),
        Book.objects.create(written=date(2018, 5, 5)),
    ]
    assert filter_field(filter_name, CO.EQ, '2019-02-12') == [books[0]]
    assert filter_field(filter_name, CO.EQ, '2019-05-14') == []
    assert filter_field(filter_name, CO.NE, '2019-02-12') == [books[1]]
    assert filter_field(filter_name, CO.LE, '2020-01-01') == books
    assert filter_field(filter_name, CO.GT, '2000-12-12') == books


@pytest.mark.django_db
def test_status():
    filter_name = 'status'
    books = [
        Book.objects.create(status=Book.WRITING),
        Book.objects.create(status=Book.PUBLISHED),
    ]
    assert filter_field(filter_name, CO.EQ, books[1].status) == [books[1]]
    assert filter_field(filter_name, CO.EQ, Book.PLANNING) == []
    assert filter_field(filter_name, CO.NE, books[1].status) == [books[0]]


@pytest.mark.django_db
def test_name():
    filter_name = 'name'
    authors = [
        Author.objects.create(name='Pushkin'),
        Author.objects.create(name='Lermontov'),
    ]
    books = [Book.objects.create(author=authors[index]) for index in range(2)]
    assert filter_field(filter_name, CO.EQ, authors[1].name) == [books[1]]
    assert filter_field(filter_name, CO.EQ, 'value__v') == []
    assert filter_field(filter_name, CO.NE, authors[1].name) == [books[0]]


@pytest.mark.django_db
@pytest.mark.parametrize('filter_name', ['author__email', 'author.email'])
def test_author__email(filter_name):
    authors = [
        Author.objects.create(email='email@gmail.com'),
        Author.objects.create(email='m.k@ingrammicro.com'),
    ]
    books = [Book.objects.create(author=authors[index]) for index in range(2)]
    assert filter_field(filter_name, CO.EQ, authors[1].email) == [books[1]]
    assert filter_field(filter_name, CO.EQ, 'email@example.com') == []
    assert filter_field(filter_name, CO.NE, authors[1].email) == [books[0]]


@pytest.mark.django_db
def test_author__is_male():
    filter_name = 'author.is_male'
    authors = [Author.objects.create(is_male=True) for _ in range(2)]
    books = [Book.objects.create(author=authors[index]) for index in range(2)]
    assert filter_field(filter_name, CO.EQ, 'true') == books
    assert filter_field(filter_name, CO.EQ, 'false') == []
    assert filter_field(filter_name, CO.NE, 'false') == books


@pytest.mark.django_db
def test_author__publisher__id():
    filter_name = 'author.publisher.id'
    publishers = [Publisher.objects.create() for _ in range(2)]
    authors = [Author.objects.create(publisher=publishers[index]) for index in range(2)]
    books = [Book.objects.create(author=authors[index]) for index in range(2)]
    assert filter_field(filter_name, CO.EQ, publishers[0].pk) == [books[0]]
    assert filter_field(filter_name, CO.EQ, 3) == []
    assert filter_field(filter_name, CO.NE, publishers[1].pk) == [books[0]]
    assert filter_field(filter_name, CO.LT, publishers[1].pk) == [books[0]]
    assert filter_field(filter_name, CO.GE, publishers[0].pk) == books


@pytest.mark.django_db
def test_page__number():
    filter_name = 'page.number'
    books = [Book.objects.create() for _ in range(2)]
    pages = [Page.objects.create(book=books[index], number=index) for index in range(2)]
    assert filter_field(filter_name, CO.EQ, pages[1].number) == [books[1]]
    assert filter_field(filter_name, CO.EQ, 22) == []
    assert filter_field(filter_name, CO.NE, pages[1].number) == [books[0]]


@pytest.mark.django_db
def test_page__id():
    filter_name = 'page.id'
    books = [Book.objects.create() for _ in range(2)]
    pages = [Page.objects.create(book=books[index]) for index in range(2)]
    assert filter_field(filter_name, CO.EQ, pages[1].pk) == [books[1]]
    assert filter_field(filter_name, CO.EQ, '5fde36e2-3442-4d2e-b221-a6758663dd72') == []
    assert filter_field(filter_name, CO.NE, pages[1].pk) == [books[0]]


@pytest.mark.django_db
def test_published_at():
    filter_name = 'published.at'
    books = [
        Book.objects.create(published_at=datetime(2019, 2, 12, 10, 2)),
        Book.objects.create(published_at=datetime(2018, 5, 5, 3, 4, 5)),
    ]
    assert filter_field(filter_name, CO.EQ, '2019-02-12T10:02:00') == [books[0]]
    assert filter_field(filter_name, CO.EQ, '2019-02-12T10:02Z') == [books[0]]
    assert filter_field(filter_name, CO.EQ, '2019-02-12T10:02:00+03:00') == []
    assert filter_field(filter_name, CO.NE, '2019-02-12T10:02') == [books[1]]
    assert filter_field(filter_name, CO.LE, '2020-01-01T00:00+08:00') == books
    assert filter_field(filter_name, CO.GT, '2000-12-12T00:21:00') == books


@pytest.mark.django_db
def test_rating_blog():
    filter_name = 'rating.blog'
    books = [
        Book.objects.create(blog_rating=Book.LOW_RATING),
        Book.objects.create(blog_rating=Book.HIGH_RATING),
    ]
    assert filter_field(filter_name, CO.EQ, Book.BLOG_RATING_CHOICES[0][1]) == [books[0]]
    assert filter_field(filter_name, CO.EQ, Book.BLOG_RATING_CHOICES[1][1]) == [books[1]]
    assert filter_field(filter_name, CO.NE, Book.BLOG_RATING_CHOICES[1][1]) == [books[0]]


@pytest.mark.django_db
def test_rating_blog_int():
    filter_name = 'rating.blog_int'
    books = [
        Book.objects.create(blog_rating=Book.LOW_RATING),
        Book.objects.create(blog_rating=Book.HIGH_RATING),
    ]
    assert filter_field(filter_name, CO.EQ, Book.LOW_RATING) == [books[0]]
    assert filter_field(filter_name, CO.EQ, Book.HIGH_RATING) == [books[1]]
    assert filter_field(filter_name, CO.NE, Book.HIGH_RATING) == [books[0]]


@pytest.mark.django_db
def test_amazon_rating():
    filter_name = 'amazon_rating'
    books = [
        Book.objects.create(amazon_rating=3.02),
        Book.objects.create(amazon_rating=2),
    ]
    assert filter_field(filter_name, CO.GE, 3.0200000) == [books[0]]
    assert filter_field(filter_name, CO.GE, 2.01) == [books[0]]
    assert filter_field(filter_name, CO.LT, 3.02) == [books[1]]


@pytest.mark.django_db
def test_url():
    filter_name = 'url'
    books = [
        Book.objects.create(publishing_url='http://www.site.com/'),
        Book.objects.create(publishing_url='https://example.com/'),
    ]
    assert filter_field(filter_name, CO.EQ, books[0].publishing_url) == [books[0]]
    assert filter_field(filter_name, CO.EQ, 'https://www.example.com/') == []
    assert filter_field(filter_name, CO.NE, books[1].publishing_url) == [books[0]]


@pytest.mark.django_db
def test_d_id():
    filter_name = 'd_id'
    authors = [Author.objects.create() for _ in range(2)]
    books = [
        Book.objects.create(author=authors[0]),
        Book.objects.create(author=authors[0]),
        Book.objects.create(author=authors[1]),
    ]
    assert filter_field(filter_name, CO.EQ, 3) == [books[2]]
    assert filter_field(filter_name, CO.EQ, 2) == [books[1], books[2]]
    assert filter_field(filter_name, CO.NE, 2) == [books[0]]
    assert filter_field(filter_name, CO.EQ, 1) == [books[0], books[1]]
    assert filter_field(filter_name, CO.NE, 1) == [books[2]]
    assert filter_field(filter_name, CO.EQ, 0) == []


@pytest.mark.django_db
def test_fsm():
    filter_name = 'fsm'
    books = [
        Book.objects.create(fsm_field=Book.STR_CHOICES.one),
        Book.objects.create(fsm_field=Book.STR_CHOICES.two),
        Book.objects.create(fsm_field=None),
    ]

    assert filter_field(filter_name, CO.EQ, Book.STR_CHOICES.one) == [books[0]]
    assert filter_field(filter_name, CO.EQ, Book.STR_CHOICES.two) == [books[1]]
    assert filter_field(filter_name, CO.NE, RQL_NULL) == [books[0], books[1]]


@pytest.mark.django_db
@pytest.mark.parametrize('filter_name', ('anno_int', 'anno_int_ref'))
def test_anno_int_ok(filter_name):
    books = create_books()

    assert filter_field(filter_name, CO.EQ, 10) == []
    assert filter_field(filter_name, CO.EQ, 1000) == books


@pytest.mark.django_db
@pytest.mark.parametrize('filter_name', ('anno_int', 'anno_int_ref'))
def test_anno_int_fail_lookup(filter_name):
    with pytest.raises(RQLFilterLookupError):
        filter_field(filter_name, SearchOperators.I_LIKE, 10)


@pytest.mark.django_db
@pytest.mark.parametrize('filter_name', ('anno_int', 'anno_int_ref'))
def test_anno_int_fail_value(filter_name):
    with pytest.raises(RQLFilterValueError):
        filter_field(filter_name, CO.EQ, 'val')


@pytest.mark.django_db
def test_anno_str_ok():
    filter_name = 'anno_str'
    books = create_books()

    assert filter_field(filter_name, CO.EQ, 'te') == []
    assert filter_field(filter_name, CO.EQ, 'text') == books


@pytest.mark.django_db
def test_anno_title_non_dynamic():
    filter_name = 'anno_title_non_dynamic'

    books = create_books()
    books[0].title = 'text'
    books[0].save(update_fields=['title'])

    assert filter_field(filter_name, CO.EQ, RQL_NULL) == [books[1]]
    assert filter_field(filter_name, CO.EQ, 'text') == [books[0]]


@pytest.mark.django_db
def test_anno_title_dynamic():
    filter_name = 'anno_title_dynamic'

    books = create_books()
    books[0].title = 'text'
    books[0].save(update_fields=['title'])

    assert filter_field(filter_name, CO.EQ, 'text') == [books[0]]
    assert filter_field(filter_name, SearchOperators.I_LIKE, 'Te*') == [books[0]]


@pytest.mark.django_db
def test_int_choice_field():
    filter_name = 'int_choice_field'
    books = [Book.objects.create(int_choice_field=choice) for choice, _ in Book.INT_CHOICES]
    assert filter_field(filter_name, CO.EQ, Book.INT_CHOICES.one) == [books[0]]
    assert filter_field(filter_name, CO.NE, Book.INT_CHOICES.two) == [books[0]]
    assert filter_field(filter_name, CO.GE, Book.INT_CHOICES.one) == books
    assert filter_field(filter_name, CO.GT, Book.INT_CHOICES.two) == []


@pytest.mark.django_db
def test_int_choice_field_repr():
    filter_name = 'int_choice_field_repr'
    books = [Book.objects.create(int_choice_field=choice) for choice, _ in Book.INT_CHOICES]
    assert filter_field(filter_name, CO.EQ, 'I') == [books[0]]
    assert filter_field(filter_name, CO.NE, 'II') == [books[0]]


@pytest.mark.django_db
def test_str_choice_field():
    filter_name = 'str_choice_field'
    books = [Book.objects.create(str_choice_field=choice) for choice, _ in Book.STR_CHOICES]
    assert filter_field(filter_name, CO.EQ, Book.STR_CHOICES.one) == [books[0]]
    assert filter_field(filter_name, CO.NE, Book.STR_CHOICES.two) == [books[0]]
    assert filter_field(filter_name, SearchOperators.I_LIKE, '*o*') == books


@pytest.mark.django_db
def test_str_choice_field_repr():
    filter_name = 'str_choice_field_repr'
    books = [Book.objects.create(str_choice_field=choice) for choice, _ in Book.STR_CHOICES]
    assert filter_field(filter_name, CO.EQ, 'I') == [books[0]]
    assert filter_field(filter_name, CO.NE, 'II') == [books[0]]


@pytest.mark.parametrize('bad_value', ['str', '2012-01-01', '2.18'])
@pytest.mark.parametrize('filter_name', ['id', 'author.publisher.id', 'page.number', 'd_id'])
def test_integer_field_fail(filter_name, bad_value):
    assert_filter_field_value_error(filter_name, CO.EQ, bad_value)


@pytest.mark.parametrize('bad_value', ['str', '2012-01-01'])
@pytest.mark.parametrize('filter_name', ['current_price', 'amazon_rating'])
def test_float_field_fail(filter_name, bad_value):
    assert_filter_field_value_error(filter_name, CO.GE, bad_value)


@pytest.mark.parametrize('bad_value', ['TRUE', '0', 'False'])
@pytest.mark.parametrize('filter_name', ['author.is_male'])
def test_boolean_field_fail(filter_name, bad_value):
    assert_filter_field_value_error(filter_name, CO.EQ, bad_value)


@pytest.mark.parametrize('bad_value', [
    '2019-02-12T10:02:00', '0', 'date', '2019:02:12', '2019-27-1',
])
@pytest.mark.parametrize('filter_name', ['written'])
def test_date_field_fail(filter_name, bad_value):
    assert_filter_field_value_error(filter_name, CO.EQ, bad_value)


@pytest.mark.parametrize('bad_value', [
    '2019-02-12', '0', 'date', '2019-02-12T27:00:00', '2019-02-12T21:00:00K',
])
@pytest.mark.parametrize('filter_name', ['published.at'])
def test_datetime_field_fail(filter_name, bad_value):
    assert_filter_field_value_error(filter_name, CO.EQ, bad_value)


@pytest.mark.parametrize('bad_operator', [CO.GT, CO.LE])
@pytest.mark.parametrize('filter_name,value', [
    ('amazon_rating', '1.23'), ('page.number', '5'),
    ('int_choice_field_repr', 'I'), ('str_choice_field_repr', 'I')
])
def test_field_lookup_fail(filter_name, value, bad_operator):
    assert_filter_field_lookup_error(filter_name, bad_operator, value)


@pytest.mark.parametrize('filter_name,bad_value', [
    ('status', 'invalid'), ('rating.blog', 'invalid'), ('rating.blog_int', '-1'),
    ('int_choice_field', 0), ('int_choice_field', 'invalid'),
    ('int_choice_field_repr', 0), ('int_choice_field_repr', 'invalid'),
    ('str_choice_field', 'zero'), ('str_choice_field_repr', 'zero'),
])
def test_bad_choice_fail(filter_name, bad_value):
    assert_filter_field_value_error(filter_name, CO.EQ, bad_value)


@pytest.mark.parametrize('bad_operator', [CO.LT, CO.GT, CO.LE, CO.LT])
@pytest.mark.parametrize('value', [RQL_NULL, RQL_EMPTY])
def test_null_empty_value_lookup_fail(bad_operator, value):
    assert_filter_field_lookup_error('title', bad_operator, value)


@pytest.mark.django_db
@pytest.mark.parametrize('operator', [CO.EQ, CO.NE, CO.GT])
@pytest.mark.parametrize('filter_name', ['invalid'])
def test_ignored_filters(filter_name, operator):
    books = create_books()
    assert filter_field(filter_name, operator, 'value') == books


@pytest.mark.parametrize('filter_name', ['id', 'page.number', 'author.is_male', 'name'])
def test_empty_value_fail(filter_name):
    assert_filter_field_value_error(filter_name, CO.EQ, RQL_EMPTY)


@pytest.mark.parametrize('value,db_lookup,db_value', [
    ('value', DjangoLookups.EXACT, 'value'),
    ('*value', DjangoLookups.ENDSWITH, 'value'),
    ('value*', DjangoLookups.STARTSWITH, 'value'),
    ('*value*', DjangoLookups.CONTAINS, 'value'),
    ('val*ue', DjangoLookups.REGEX, '^val(.*?)ue$'),
    ('val*ue*', DjangoLookups.REGEX, '^val(.*?)ue'),
    ('*val*ue', DjangoLookups.REGEX, 'val(.*?)ue$'),
    ('*val*ue*', DjangoLookups.REGEX, 'val(.*?)ue'),
    ('*', DjangoLookups.REGEX, '(.*?)'),
    ('value\*', DjangoLookups.EXACT, 'value*'),
    (r'value\\*', DjangoLookups.STARTSWITH, 'value\\'),
    (r'value\\\*', DjangoLookups.EXACT, r'value\*'),
    (r'value\**', DjangoLookups.STARTSWITH, 'value*'),
    ('*\*\*value', DjangoLookups.ENDSWITH, '**value'),
    ('*val\*ue*', DjangoLookups.CONTAINS, 'val*ue'),
    ('va\*l*\*ue*', DjangoLookups.REGEX, '^va*l(.*?)*ue'),
])
def test_searching_q_ok(value, db_lookup, db_value):
    cls = BooksFilterClass(book_qs)

    for v in (value, '"{}"'.format(value)):
        like_q = cls.build_q_for_filter(FilterArgs('title', SearchOperators.LIKE, v))
        assert like_q.children[0] == ('title__{}'.format(db_lookup), db_value)

    i_like_q = cls.build_q_for_filter(FilterArgs('title', SearchOperators.I_LIKE, value))
    assert i_like_q.children[0] == ('title__i{}'.format(db_lookup), db_value)


@pytest.mark.django_db
def test_searching_db_ok():
    filter_name = 'title'
    title = 'Long title'
    book = Book.objects.create(title=title)
    assert filter_field(filter_name, SearchOperators.I_LIKE, title.upper()) == [book]
    assert filter_field(filter_name, SearchOperators.LIKE, 'title') == []
    assert filter_field(filter_name, SearchOperators.LIKE, 'L*') == [book]
    assert filter_field(filter_name, SearchOperators.I_LIKE, '*E') == [book]
    assert filter_field(filter_name, SearchOperators.LIKE, '*ng*') == [book]
    assert filter_field(filter_name, SearchOperators.LIKE, '*ng*') == [book]
    assert filter_field(filter_name, SearchOperators.LIKE, '*t*t*') == [book]
    assert filter_field(filter_name, SearchOperators.LIKE, '*t*le') == [book]
    assert filter_field(filter_name, SearchOperators.I_LIKE, 'lo*g*') == [book]
    assert filter_field(filter_name, SearchOperators.LIKE, 't*le') == []
    assert filter_field(filter_name, SearchOperators.LIKE, '*') == [book]


@pytest.mark.parametrize('bad_value', ['value**value', '**v'])
@pytest.mark.parametrize('operator', [SearchOperators.LIKE, SearchOperators.I_LIKE])
def test_searching_value_fail(bad_value, operator):
    assert_filter_field_value_error('title', operator, bad_value)


@pytest.mark.django_db
def test_custom_filter_ok():
    class CustomCls(BooksFilterClass):
        def build_q_for_custom_filter(self, *args, **kwargs):
            return Q(id__gte=2)

    filter_cls = CustomCls(book_qs)
    q = filter_cls.build_q_for_filter(FilterArgs('custom_filter', SearchOperators.I_LIKE, 'value'))

    books = [Book.objects.create() for _ in range(2)]
    assert list(book_qs.filter(q)) == [books[1]]


def test_custom_filter_fail():
    with pytest.raises(RQLFilterParsingError) as e:
        filter_field('custom_filter', SearchOperators.I_LIKE, 'value')
    assert e.value.details['error'] == 'Filter logic is not implemented: custom_filter.'


def test_custom_filter_ordering_fail():
    with pytest.raises(RQLFilterParsingError) as e:
        BooksFilterClass(book_qs).build_name_for_custom_ordering('custom_filter')
    assert e.value.details['error'] == 'Ordering logic is not implemented: custom_filter.'



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from functools import partial

import pytest
from django.core.exceptions import FieldError
from django.db.models import Q, IntegerField
from django.utils.timezone import now

from dj_rql.constants import FilterLookups, ListOperators, RQL_NULL
from dj_rql.exceptions import RQLFilterLookupError, RQLFilterParsingError
from dj_rql.filter_cls import RQLFilterClass
from tests.dj_rf.filters import BooksFilterClass
from tests.dj_rf.models import Author, Book, Publisher
from tests.test_filter_cls.utils import book_qs, create_books


def apply_filters(query):
    filter_cls = BooksFilterClass(book_qs)
    _, q = filter_cls.apply_filters(query)
    return list(q)


def test_parsing_error():
    bad_query = 'q='
    with pytest.raises(RQLFilterParsingError) as e:
        apply_filters(bad_query)
    assert e.value.details['error'] == 'Bad filter query.'


def test_lookup_error():
    bad_lookup = 'like(id,1)'
    with pytest.raises(RQLFilterLookupError):
        apply_filters(bad_lookup)


@pytest.mark.django_db
@pytest.mark.parametrize('comparison_tpl', ['eq(title,{})', 'title=eq={}', 'title={}'])
def test_comparison(comparison_tpl):
    title = 'book'
    books = [
        Book.objects.create(title=title),
        Book.objects.create(title='another'),
    ]
    assert apply_filters(comparison_tpl.format(title)) == [books[0]]


@pytest.mark.django_db
@pytest.mark.parametrize('searching_tpl', ['like(title,*{}*)', 'ilike(title,*{})'])
def test_searching(searching_tpl):
    title = 'book'
    books = [
        Book.objects.create(title=title),
        Book.objects.create(title='another'),
    ]
    assert apply_filters(searching_tpl.format(title)) == [books[0]]


@pytest.mark.django_db
@pytest.mark.parametrize('operator', ['&', ','])
def test_and(operator):
    email, title = 'george@martin.com', 'book'
    comp1 = 'title={}'.format(title)
    comp2 = 'eq(author.email,{})'.format(email)
    query = '{comp1}{op}{comp2}'.format(comp1=comp1, op=operator, comp2=comp2)

    authors = [
        Author.objects.create(email='email@example.com'),
        Author.objects.create(email=email),
    ]
    books = [Book.objects.create(author=authors[index], title=title) for index in range(2)]

    expected = [books[1]]
    assert apply_filters(query) == expected
    assert apply_filters('{q}{op}{q}'.format(q=query, op=operator)) == expected
    assert apply_filters('and({comp1},{comp2})'.format(comp1=comp1, comp2=comp2)) == expected


@pytest.mark.django_db
@pytest.mark.parametrize('operator', ['|', ';'])
def test_or(operator):
    email, title = 'george@martin.com', 'book'
    comp1 = 'title={}'.format(title)
    comp2 = 'eq(author.email,{})'.format(email)
    query = '({comp1}{op}{comp2})'.format(comp1=comp1, op=operator, comp2=comp2)

    authors = [
        Author.objects.create(email='email@example.com'),
        Author.objects.create(email=email),
    ]
    books = [Book.objects.create(author=authors[index], title=title) for index in range(2)]

    expected = books
    assert apply_filters(query) == expected
    assert apply_filters('or({comp1},{comp2})'.format(comp1=comp1, comp2=comp2)) == expected


@pytest.mark.django_db
def test_not():
    title = 'book'
    books = [
        Book.objects.create(title=title),
        Book.objects.create(title='another'),
    ]
    assert apply_filters('not(title={})'.format(title)) == [books[1]]


@pytest.mark.django_db
def test_nested_logic():
    publisher = [Publisher.objects.create() for _ in range(2)]

    author = Author.objects.create(name='Foo', publisher=publisher[0], is_male=False)
    book = Book.objects.create(amazon_rating=4.0, author=author)

    other_author = Author.objects.create(name='Bar', publisher=publisher[0], is_male=False)
    other_book = Book.objects.create(amazon_rating=4.5, author=other_author)

    assert apply_filters('') == [book, other_book]
    assert apply_filters(
        '(author.publisher.id=2|ne(name,"Bar")),'
        '(eq(author.is_male,true);amazon_rating=ge=4.000)',
    ) == [book]
    assert apply_filters(
        '(and(author.is_male=ne=true,author.is_male=true);gt(id,1))',
    ) == [other_book]


def apply_listing_filters(operator, *values):
    return apply_filters('{operator}(id,({values}))'.format(
        operator=operator, values=','.join(tuple(values)),
    ))


apply_in_listing_filters = partial(apply_listing_filters, ListOperators.IN)
apply_out_listing_filters = partial(apply_listing_filters, ListOperators.OUT)


@pytest.mark.django_db
def test_in():
    books = create_books()
    assert apply_in_listing_filters(str(books[0].pk)) == [books[0]]
    assert apply_in_listing_filters(str(books[1].pk), '23') == [books[1]]
    assert apply_in_listing_filters(str(books[1].pk), str(books[0].pk)) == books
    assert apply_in_listing_filters('23') == []


@pytest.mark.django_db
def test_out():
    books = create_books()
    assert apply_out_listing_filters(str(books[0].pk)) == [books[1]]
    assert apply_out_listing_filters(str(books[1].pk), '23') == [books[0]]
    assert apply_out_listing_filters(str(books[1].pk), str(books[0].pk)) == []
    assert apply_out_listing_filters('23') == books


@pytest.mark.django_db
def test_null():
    books = create_books()
    assert apply_filters('title={}'.format(RQL_NULL)) == books
    assert apply_filters('title=ne={}'.format(RQL_NULL)) == []


@pytest.mark.django_db
def test_null_with_in_or():
    books = create_books()

    title = 'null'
    books[0].title = title
    books[0].save(update_fields=['title'])

    assert apply_filters('in(title,({},{}))'.format(title, RQL_NULL)) == books
    assert apply_filters('or(title=eq={},eq(title,{}))'.format(title, RQL_NULL)) == books


@pytest.mark.django_db
def test_null_on_foreign_key_pk():
    publisher = Publisher.objects.create()
    authors = [
        Author.objects.create(email='a@m.com', publisher=publisher),
        Author.objects.create(email='z@m.com'),
    ]
    books = [Book.objects.create(author=author) for author in authors]

    assert apply_filters('author.publisher.id={}'.format(RQL_NULL)) == [books[1]]
    assert apply_filters('ne(author.publisher.id,{})'.format(RQL_NULL)) == [books[0]]


@pytest.mark.django_db
def test_ordering_source():
    authors = [
        Author.objects.create(email='a@m.com'),
        Author.objects.create(email='z@m.com'),
    ]
    books = [Book.objects.create(author=author) for author in authors]
    assert apply_filters('ordering(-author.email)') == [books[1], books[0]]
    assert apply_filters('ordering(author.email)') == [books[0], books[1]]
    assert apply_filters('ordering(author.email,-author.email)') == [books[0], books[1]]


@pytest.mark.django_db
def test_ordering_sources():
    books = create_books()
    assert apply_filters('ordering(d_id)') == [books[0], books[1]]
    assert apply_filters('ordering(-d_id)') == [books[1], books[0]]


@pytest.mark.django_db
def test_ordering_by_several_filters():
    same_email = 'a@m.com'
    authors = [
        Author.objects.create(email=same_email),
        Author.objects.create(email=same_email),
    ]
    books = [Book.objects.create(author=author, published_at=now()) for author in authors]
    books.append(Book.objects.create(published_at=now()))

    assert apply_filters('ordering(author.email,-published.at)') == \
        list(book_qs.order_by('author__email', '-published_at'))


@pytest.mark.django_db
def test_ordering_by_empty_value():
    books = create_books()
    assert apply_filters('ordering()') == books


def test_several_ordering_operations():
    with pytest.raises(RQLFilterParsingError) as e:
        apply_filters('ordering(d_id)&ordering(author.email)')
    assert e.value.details['error'] == \
        'Bad ordering filter: query can contain only one ordering operation.'


def test_bad_ordering_filter():
    with pytest.raises(RQLFilterParsingError) as e:
        apply_filters('ordering(id)')
    assert e.value.details['error'] == 'Bad ordering filter: id.'


@pytest.mark.django_db
def test_search():
    title = 'book'
    books = [
        Book.objects.create(title=title),
        Book.objects.create(title='another'),
    ]

    assert apply_filters('(search=*aN*r;search=b*k)') == books
    assert apply_filters('(search=Ano*;search=oO)') == books
    assert apply_filters('(search=*aN*r;search=b*k)') == books
    assert apply_filters('(search="*aN*";search="*book")') == books
    assert apply_filters('(search="*aN";search="*b*")') == books
    assert apply_filters('search=bo') == [books[0]]


def test_search_bad_lookup():
    with pytest.raises(RQLFilterParsingError) as e:
        apply_filters('search=ge=*a*')
    assert e.value.details['error'] == 'Bad search filter: ge.'


@pytest.mark.django_db
@pytest.mark.parametrize('operator', (ListOperators.IN, ListOperators.OUT))
def test_custom_filter_list_lookup_ok(operator):
    class CustomCls(BooksFilterClass):
        def build_q_for_custom_filter(self, *args, **kwargs):
            return Q(id__gte=2)

    books = [Book.objects.create() for _ in range(2)]
    assert list(
        CustomCls(book_qs).apply_filters('{}(has_list_lookup,(1,2))'.format(operator))[1]
    ) == [books[1]]


@pytest.mark.parametrize('operator', (ListOperators.IN, ListOperators.OUT))
def test_custom_filter_list_lookup_fail(operator):
    class CustomCls(BooksFilterClass):
        def build_q_for_custom_filter(self, *args, **kwargs):
            return Q(id__gte=2)

    with pytest.raises(RQLFilterLookupError) as e:
        CustomCls(book_qs).apply_filters('{}(no_list_lookup,(1,2))'.format(operator))
    assert e.value.details['lookup'] == operator


@pytest.mark.django_db
def test_custom_filter_ordering():
    class CustomCls(BooksFilterClass):
        def build_name_for_custom_ordering(self, filter_name):
            return 'id'

        def assert_ordering(self, filter_name, expected):
            assert list(self.apply_filters('ordering({})'.format(filter_name))[1]) == expected

    books = create_books()

    CustomCls(book_qs).assert_ordering('ordering_filter', [books[0], books[1]])
    CustomCls(book_qs).assert_ordering('-ordering_filter', [books[1], books[0]])


@pytest.mark.django_db
def test_custom_filter_search_ok(mocker):
    class CustomCls(RQLFilterClass):
        MODEL = Book
        FILTERS = [{
            'filter': 'search_filter',
            'custom': True,
            'search': True,
            'lookups': {FilterLookups.I_LIKE},
        }]

        def assert_search(self, value, expected):
            assert list(self.apply_filters('search={}'.format(value))[1]) == expected

        @classmethod
        def side_effect(cls, data):
            django_lookup = data.django_lookup
            return Q(**{
                'title__{}'.format(django_lookup): cls._get_searching_typed_value(
                    django_lookup, data.str_value,
                )
            })

    build_q_for_custom_filter_patch = mocker.patch.object(
        CustomCls, 'build_q_for_custom_filter', side_effect=CustomCls.side_effect,
    )

    books = [
        Book.objects.create(title='book'),
        Book.objects.create(title='another'),
    ]

    CustomCls(book_qs).assert_search('ok', [books[0]])
    CustomCls(book_qs).assert_search('AN', [books[1]])
    CustomCls(book_qs).assert_search('o', books)

    assert build_q_for_custom_filter_patch.call_count == 3


@pytest.mark.django_db
def test_dynamic_no_annotation():
    class CustomCls(RQLFilterClass):
        MODEL = Book
        FILTERS = [{
            'filter': 'anno',
            'dynamic': True,
            'field': IntegerField(),
        }]

    # We want to be error unhandled in this case
    with pytest.raises(FieldError):
        CustomCls(book_qs).apply_filters('anno=5')


@pytest.mark.django_db
def test_extended_search_ok():
    class CustomCls(RQLFilterClass):
        MODEL = Book
        FILTERS = ['id']
        EXTENDED_SEARCH_ORM_ROUTES = ['title']

        def assert_search(self, value, expected):
            assert list(self.apply_filters('search={}'.format(value))[1]) == expected

    books = [
        Book.objects.create(title='book'),
        Book.objects.create(title='another'),
    ]

    CustomCls(book_qs).assert_search('ok', [books[0]])
    CustomCls(book_qs).assert_search('AN', [books[1]])
    CustomCls(book_qs).assert_search('o', books)


@pytest.mark.django_db
def test_extended_search_fail():
    class CustomCls(RQLFilterClass):
        MODEL = Book
        FILTERS = ['id']
        EXTENDED_SEARCH_ORM_ROUTES = ['invalid']

    # We want to be error unhandled in this case
    with pytest.raises(FieldError):
        CustomCls(book_qs).apply_filters('search=text')


@pytest.mark.django_db
@pytest.mark.parametrize('select', ('title,+page,-rating.blog_int', '', 'title'))
def test_select(select):
    assert apply_filters('select({})'.format(select)) == []


@pytest.mark.django_db
def test_braces_in_braces():
    books = [
        Book.objects.create(title='book'),
        Book.objects.create(title='another'),
    ]
    assert apply_filters('(((title=book)))&(title=book)') == [books[0]]
    assert apply_filters('(title=book|(title=invalid))') == [books[0]]


@pytest.mark.django_db
@pytest.mark.parametrize('distinct', (True, False))
def test_global_distinct(distinct):
    class CustomCls(RQLFilterClass):
        MODEL = Book
        FILTERS = ['id']
        DISTINCT = distinct

    _, qs = CustomCls(book_qs).apply_filters('id=1')
    assert qs.query.distinct is distinct


@pytest.mark.django_db
def test_distinct_on_field_no_field_in_filter():
    _, qs = BooksFilterClass(book_qs).apply_filters('title=abc')
    assert not qs.query.distinct


@pytest.mark.django_db
def test_distinct_on_field_field_in_filter():
    _, qs = BooksFilterClass(book_qs).apply_filters('status=planning')
    assert qs.query.distinct


@pytest.mark.django_db
def test_two_distinct_fields():
    _, qs = BooksFilterClass(book_qs).apply_filters('status=planning,name=author')
    assert qs.query.distinct


@pytest.mark.django_db
def test_distinct_with_custom():
    class CustomCls(BooksFilterClass):
        def build_q_for_custom_filter(self, *args, **kwargs):
            return Q()

    _, qs = CustomCls(book_qs).apply_filters('ilike(custom_filter,text)')
    assert qs.query.distinct


@pytest.mark.django_db
def test_distinct_on_field_field_in_ordering():
    _, qs = BooksFilterClass(book_qs).apply_filters('ordering(published.at)')
    assert qs.query.distinct


@pytest.mark.django_db
def test_distinct_on_field_field_not_in_ordering():
    _, qs = BooksFilterClass(book_qs).apply_filters('ordering(int_choice_field)')
    assert not qs.query.distinct



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest

from dj_rql.exceptions import (
    RQLFilterError, RQLFilterLookupError, RQLFilterParsingError, RQLFilterValueError,
)


@pytest.mark.parametrize('exception_cls,message', [
    (RQLFilterError, 'RQL Filtering error.'),
    (RQLFilterLookupError, 'RQL Lookup error.'),
    (RQLFilterParsingError, 'RQL Parsing error.'),
    (RQLFilterValueError, 'RQL Value error.'),
])
def test_exception_message(exception_cls, message):
    with pytest.raises(exception_cls) as e:
        raise exception_cls
    assert str(e.value) == message



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest
from lark.exceptions import LarkError

from dj_rql.constants import RQL_ORDERING_OPERATOR
from dj_rql.parser import RQLParser
from tests.test_parser.constants import FAIL_PROPS, OK_PROPS
from tests.test_parser.utils import OrderingTransformer


REVERSED_OK_PROPS = reversed(OK_PROPS)


def ordering_transform(props):
    query = '{operator}({props})'.format(
        operator=RQL_ORDERING_OPERATOR, props=','.join(props),
    )
    return OrderingTransformer().transform(RQLParser.parse(query))


@pytest.mark.parametrize('p1,p2', zip(OK_PROPS, REVERSED_OK_PROPS))
def test_ordering_ok(p1, p2):
    assert ordering_transform(('+{}'.format(p1),)) == (p1,)
    assert ordering_transform(('-{}'.format(p2),)) == ('-{}'.format(p2),)
    assert ordering_transform((p2, p1)) == (p2, p1)
    assert ordering_transform((p2, p1, '+{}'.format(p2))) == (p2, p1, p2)


def test_ordering_empty_ok():
    assert ordering_transform([]) == ()


@pytest.mark.parametrize('prop', FAIL_PROPS[1:])
def test_ordering_property_fail(prop):
    with pytest.raises(LarkError):
        ordering_transform(prop)


# Temporary is here
@pytest.mark.parametrize('prop', ('', 'prop'))
def test_select_ok(prop):
    RQLParser.parse('select({})'.format(prop))



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from collections import OrderedDict

import pytest

from tests.dj_rf.models import Author, Book, Page, Publisher
from tests.dj_rf.serializers import SelectBookSerializer


@pytest.mark.django_db
def test_select_complex():
    publisher = Publisher.objects.create(name='publisher')
    author = Author.objects.create(name='auth', publisher=publisher)
    book = Book.objects.create(author=author)
    page = Page.objects.create(book=book, number=1, content='text')

    select = OrderedDict()
    select['blog_rating'] = True
    select['github_stars'] = False
    select['author'] = True
    select['author.publisher'] = True
    select['author.publisher.name'] = False
    select['author_ref.name'] = False
    select['pages'] = True
    select['pages.content'] = False

    class Request:
        rql_select = {
            'depth': 0,
            'select': select,
        }

    data = SelectBookSerializer(book, context={'request': Request}).data
    assert {
        'id': book.id,
        'blog_rating': None,
        'star': {},
        'author_ref': {
            'id': author.id,
        },
        'author': {
            'id': author.id,
            'name': 'auth',
            'publisher': {
                'id': publisher.id,
            },
        },
        'pages': [{
            'id': str(page.uuid),
        }]} == data


@pytest.mark.django_db
def test_select_request_without_rql_select():
    book = Book.objects.create()

    class Request:
        pass

    data = SelectBookSerializer(book, context={'request': Request}).data
    assert data


@pytest.mark.django_db
def test_select_no_request():
    book = Book.objects.create()

    data = SelectBookSerializer(book).data
    assert data


@pytest.mark.django_db
def test_select_misconfiguration():
    book = Book.objects.create()

    select = OrderedDict()
    select['invalid'] = True
    select['invalid.invalid'] = False

    class Request:
        rql_select = {
            'depth': 0,
            'select': select,
        }

    data = SelectBookSerializer(book, context={'request': Request}).data
    assert data



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest

from dj_rql.utils import assert_filter_cls
from tests.data import get_book_filter_cls_ordering_data, get_book_filter_cls_search_data
from tests.dj_rf.filters import BooksFilterClass


def test_ordering_assertion():
    with pytest.raises(AssertionError) as e:
        assert_filter_cls(BooksFilterClass, {}, set(), get_book_filter_cls_search_data())

    assert str(e.value) == "Ordering filter data doesn't match."


def test_search_assertion():
    with pytest.raises(AssertionError) as e:
        assert_filter_cls(BooksFilterClass, {}, get_book_filter_cls_ordering_data(), set())

    assert str(e.value) == "Searching filter data doesn't match."


def test_no_key():
    mismatch = {'key': {'orm_route': 'value'}}

    with pytest.raises(AssertionError) as e:
        assert_filter_cls(BooksFilterClass, mismatch, set(), set())

    assert "Filter `key` is not set" in str(e.value)


def test_mismatch_for_list_data():
    mismatch = {
        'd_id': [
            {'orm_route': 'id'},
            {'orm_route': 'author__id'},
            {'orm_route': 'invalid'},
        ],
    }

    with pytest.raises(AssertionError) as e:
        assert_filter_cls(BooksFilterClass, mismatch, set(), set())

    assert "Filter `d_id` data doesn't match" in str(e.value)


def test_mismatch_for_value():
    mismatch = {'id': {'orm_route': 'invalid'}}

    with pytest.raises(AssertionError) as e:
        assert_filter_cls(BooksFilterClass, mismatch, set(), set())

    assert str(e.value) == 'Wrong filter `id` configuration: id != invalid'


def test_mismatch_for_fields():
    mismatch = {'id': {'orm_route': 'id'}}

    with pytest.raises(AssertionError) as e:
        assert_filter_cls(BooksFilterClass, mismatch, set(), set())

    assert str(e.value) == "Wrong filter `id` configuration: assertion data " \
                           "must contain `orm_route` and `lookups`."



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from rest_framework import serializers

from dj_rql.drf.serializers import RQLMixin
from tests.dj_rf.models import Publisher, Author, Page, Book


class PublisherReferenceSerializer(RQLMixin, serializers.ModelSerializer):
    class Meta:
        model = Publisher
        fields = ('id', 'name')


class AuthorReferenceSerializer(RQLMixin, serializers.ModelSerializer):
    class Meta:
        model = Author
        fields = ('id', 'name')


class AuthorSerializer(RQLMixin, serializers.ModelSerializer):
    publisher = PublisherReferenceSerializer()

    class Meta:
        model = Author
        fields = (
            'id',
            'name',
            'publisher',
        )


class PageReferenceSerializer(RQLMixin, serializers.ModelSerializer):
    id = serializers.CharField(source='uuid')

    class Meta:
        model = Page
        fields = (
            'id',
            'content',
        )


class StarAuthorSerializer(serializers.Serializer):
    def to_representation(self, instance):
        return {}


class SelectBookSerializer(RQLMixin, serializers.ModelSerializer):
    write_only = serializers.CharField(source='title', write_only=True)

    author_ref = AuthorReferenceSerializer(source='author')
    author = serializers.SerializerMethodField()
    pages = PageReferenceSerializer(many=True)
    star = StarAuthorSerializer(source='*')

    class Meta:
        model = Book
        fields = (
            'id',  # must always be shown in both detail and list
            'blog_rating',  # by default not shown in list, but shown in detail
            'github_stars',  # by default not shown anywhere
            'star',
            'write_only',
            'author_ref',  # One level reference field (FK)
            'author',  # Deep nested fields (FK)
            'pages',  # List of backrefs
        )

    def get_author(self, obj):
        if obj.author:
            return AuthorSerializer(obj.author, context=self.rql_context('author')).data


class BookSerializer(serializers.ModelSerializer):
    class Meta:
        model = Book
        fields = ('id',)



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest
from django.http import QueryDict
from django.utils.timezone import now
from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK

from dj_rql.constants import DjangoLookups
from dj_rql.drf.compat import CompatibilityRQLFilterBackend, DjangoFiltersRQLFilterBackend
from dj_rql.exceptions import RQLFilterParsingError, RQLFilterValueError
from tests.dj_rf.filters import BooksFilterClass
from tests.dj_rf.models import Author, Book


def test_compatibility_is_old_syntax():
    with pytest.raises(NotImplementedError):
        CompatibilityRQLFilterBackend.is_old_syntax(None, None, None)


def test_compatibility_get_rql_query():
    with pytest.raises(NotImplementedError):
        CompatibilityRQLFilterBackend.get_rql_query(None, None, None)


@pytest.mark.parametrize('backend', (CompatibilityRQLFilterBackend, DjangoFiltersRQLFilterBackend))
def test_compatibility_modify_initial_query(backend):
    assert backend.modify_initial_query(None, None, 'q') == 'q'


@pytest.mark.parametrize('query,expected', (
    ('', False),
    ('&', True),
    ('k=v&&', True),
    ('k=True', True),
    ('author.is_male=True', True),
    ('&k&', True),
    ('k=v', False),
    ('order_by=v', True),
    ('k=v%20v', True),
    ('k__in=v', True),
    ('l1__l2__in=v', True),
    ('k__random=v', False),
    ('k__in=v,v', True),
    ('k__in=ne', True),
    ('k__in=ne=v', False),
    ('k__in=ne=v""', False),
    ('t__in=v', False),
    # TODO: Check ('title__in=v,v', True),
    ('order_by=k&k__in=v', True),
    ('limit=10,offset=2', False),
    ('limit=10,eq(offset,2)', False),
    ('limit=10,eq(offset__in,2)', False),
    ('limit=10,eq(t__in,b)', False),
    ('limit=10;k__in=2', True),
    ('limit=10;eq(t__in,b)', False),
    ('select(books)&k__in=v,v', True),
    ('k__in=v,v&select(books)', True),
    ('select(books)', False),
))
def test_old_syntax(mocker, query, expected):
    request = mocker.MagicMock(query_params=QueryDict(query))
    filter_instance = BooksFilterClass(Book.objects.none())
    assert DjangoFiltersRQLFilterBackend.is_old_syntax(filter_instance, request, query) == expected


def test_old_syntax_filters(mocker):
    query = 'k__in=v'
    request = mocker.MagicMock(query_params=QueryDict(query))
    filter_instance = BooksFilterClass(Book.objects.none())

    for _ in range(2):
        assert DjangoFiltersRQLFilterBackend.is_old_syntax(filter_instance, request, query) is True
        assert filter_instance.old_syntax_filters == {'t__in'}


@pytest.mark.parametrize('query', (
    'select(books)&k__in=v,v', 'k__in=v,v&select(books)',
))
def test_old_syntax_select_remains(query, mocker):
    expected = ('select(books)&in(k,("v","v"))', 'in(k,("v","v"))&select(books)')
    request = mocker.MagicMock(query_params=QueryDict(query))
    filter_instance = BooksFilterClass(Book.objects.none())
    assert DjangoFiltersRQLFilterBackend.get_rql_query(filter_instance, request, query) in expected


def filter_api(api_client, query):
    return api_client.get('{}?{}'.format(reverse('old_book-list'), query))


def assert_ok_response(response, count):
    assert response.status_code == HTTP_200_OK
    assert len(response.data) == count


def create_book(title=None):
    return Book.objects.create(title=title)


def create_books(titles):
    return [create_book(title=title) for title in titles]


@pytest.mark.django_db
def test_common_comparison(api_client, clear_cache):
    books = create_books(['G', 'H'])

    response = filter_api(api_client, 'order_by=published.at&title=G')
    assert_ok_response(response, 1)
    assert response.data[0]['id'] == books[0].id


@pytest.mark.django_db
@pytest.mark.parametrize('title,count', (
    ('G', 1),
    ('', 2),
))
def test__in_with_one_value(api_client, clear_cache, title, count):
    create_books(['G', ''])

    response = filter_api(api_client, 'title__in={}'.format(title))
    assert_ok_response(response, count)


@pytest.mark.django_db
def test__in_with_several_values(api_client, clear_cache):
    books = create_books(['G', 'H', ''])

    response = filter_api(api_client, 'title__in=G,H')
    assert_ok_response(response, 2)
    assert [response.data[0]['id'], response.data[1]['id']] == [books[0].id, books[1].id]


@pytest.mark.django_db
@pytest.mark.parametrize('query', (
    'G,',
    ',G',
))
def test__in_with_several_values_one_empty_one_invalid(api_client, clear_cache, query):
    create_books(['G', ''])

    response = filter_api(api_client, 'title__in={}'.format(query))
    assert_ok_response(response, 1)


@pytest.mark.django_db
def test_double__in_property(api_client, clear_cache):
    books = create_books(['G', 'H', ''])

    response = filter_api(api_client, 'title__in=G,H&title__in=,G')
    assert_ok_response(response, 1)
    assert response.data[0]['id'] == books[0].id


@pytest.mark.django_db
@pytest.mark.parametrize('value,index', (
    ('True', 0),
    ('true', 0),
    ('False', 1),
    ('false', 1),
))
def test_boolean_value_ok(api_client, clear_cache, value, index):
    authors = [
        Author.objects.create(name='n', is_male=True),
        Author.objects.create(name='n', is_male=False),
    ]
    books = [Book.objects.create(author=author) for author in authors]

    response = filter_api(api_client, 'author.is_male={}'.format(value))
    assert_ok_response(response, 1)
    assert response.data[0]['id'] == books[index].id


@pytest.mark.django_db
@pytest.mark.parametrize('value', (
    '0',
    '1',
    'TRUE',
    'other',
))
def test_boolean_value_fail(api_client, clear_cache, value):
    with pytest.raises(RQLFilterValueError):
        filter_api(api_client, 'author.is_male={}'.format(value))


@pytest.mark.django_db
@pytest.mark.parametrize('value,index', (
    ('True', 0),
    ('true', 0),
    ('1', 0),
    ('False', 1),
    ('false', 1),
    ('0', 1),
))
def test__isnull_ok(api_client, clear_cache, value, index):
    books = create_books([None, 'G'])

    response = filter_api(api_client, 'title__isnull={}'.format(value))
    assert_ok_response(response, 1)
    assert response.data[0]['id'] == books[index].id


@pytest.mark.django_db
@pytest.mark.parametrize('value', (
    '2',
    'TRUE',
    'other',
))
def test__isnull_fail(api_client, clear_cache, value):
    with pytest.raises(RQLFilterParsingError):
        filter_api(api_client, 'title__isnull={}'.format(value))


@pytest.mark.django_db
@pytest.mark.parametrize('value,count', (
    ('"G"', 0),
    ('g', 0),
    ('G', 1),
))
def test__exact(api_client, clear_cache, value, count):
    create_book('G')

    response = filter_api(api_client, 'title__exact={}'.format(value))
    assert_ok_response(response, count)


@pytest.mark.django_db
@pytest.mark.parametrize('values,count', (
    (('G', 'G'), 1),
    (('G', 'H'), 0),
))
def test_multiple_choice(api_client, clear_cache, values, count):
    create_book('G')

    response = filter_api(
        api_client, 'title__exact={}&title__exact={}'.format(values[0], values[1]),
    )
    assert_ok_response(response, count)


@pytest.mark.django_db
@pytest.mark.parametrize('operator', (
    DjangoLookups.CONTAINS,
    DjangoLookups.I_CONTAINS,
))
@pytest.mark.parametrize('value,count', (
    ('Title', 1),
    ('Ti', 1),
    ('it', 1),
    ('le', 1),
    ('"Title"', 0),
    ('other', 0),
    ('iT', 1),
    ('titlE', 1),
))
def test__contains(api_client, clear_cache, value, count, operator):
    create_book('Title')

    response = filter_api(api_client, 'title__{}={}'.format(operator, value))
    assert_ok_response(response, count)


@pytest.mark.django_db
@pytest.mark.parametrize('operator', (
    DjangoLookups.STARTSWITH,
    DjangoLookups.I_STARTSWITH,
))
@pytest.mark.parametrize('value,count', (
    ('Title', 1),
    ('Ti', 1),
    ('it', 0),
    ('le', 0),
    ('"Title"', 0),
    ('other', 0),
    ('iT', 0),
    ('titlE', 1),
))
def test__startswith(api_client, clear_cache, value, count, operator):
    create_book('Title')

    response = filter_api(api_client, 'title__{}={}'.format(operator, value))
    assert_ok_response(response, count)


@pytest.mark.django_db
@pytest.mark.parametrize('operator', (
    DjangoLookups.ENDSWITH,
    DjangoLookups.I_ENDSWITH,
))
@pytest.mark.parametrize('value,count', (
    ('Title', 1),
    ('Ti', 0),
    ('it', 0),
    ('le', 1),
    ('"Title"', 0),
    ('other', 0),
    ('iT', 0),
    ('titlE', 1),
))
def test__endswith(api_client, clear_cache, value, count, operator):
    create_book('Title')

    response = filter_api(api_client, 'title__{}={}'.format(operator, value))
    assert_ok_response(response, count)


@pytest.mark.django_db
@pytest.mark.parametrize('lookup', (
    DjangoLookups.REGEX,
    DjangoLookups.I_REGEX,
))
def test__regex(api_client, clear_cache, lookup):
    with pytest.raises(RQLFilterParsingError):
        filter_api(api_client, 'title__{}=true'.format(lookup))


@pytest.mark.django_db
@pytest.mark.parametrize('lookup', (
    'day',
    'week',
    'month',
    'time',
    'hour',
    'minute',
    'second',
    'day__gt',
))
def test__day_week_etc(api_client, clear_cache, lookup):
    response = filter_api(api_client, 'title__{}=2020'.format(lookup))
    assert_ok_response(response, 0)


@pytest.mark.django_db
@pytest.mark.parametrize('lookup,p_value,n_value', (
    (DjangoLookups.GT, 1, 5),
    (DjangoLookups.GTE, 5, 10),
    (DjangoLookups.LT, 10, 5),
    (DjangoLookups.LTE, 10, 1),
))
def test__gt_ge_lt_le(api_client, clear_cache, lookup, p_value, n_value):
    Book.objects.create(github_stars=5)

    response = filter_api(api_client, 'github_stars__{}={}'.format(lookup, p_value))
    assert_ok_response(response, 1)

    response = filter_api(api_client, 'github_stars__{}={}'.format(lookup, n_value))
    assert_ok_response(response, 0)


@pytest.mark.django_db
@pytest.mark.parametrize('ordering_term', (
    'order_by',
    'ordering',
))
def test_order_ok(api_client, clear_cache, ordering_term):
    same_email = 'a@m.com'
    authors = [
        Author.objects.create(email=same_email),
        Author.objects.create(email=same_email),
    ]
    books = [Book.objects.create(author=author, published_at=now()) for author in authors]
    books.append(Book.objects.create(published_at=now()))

    response = filter_api(api_client, '{}=author.email,-published.at'.format(ordering_term))
    assert [d['id'] for d in response.data] == \
        list(b.id for b in Book.objects.all().order_by('author__email', '-published_at'))


@pytest.mark.django_db
@pytest.mark.parametrize('ordering_term', (
    'order_by',
    'ordering',
))
def test_order_fail(api_client, clear_cache, ordering_term):
    with pytest.raises(RQLFilterParsingError):
        filter_api(api_client, '{}=invalid'.format(ordering_term))


@pytest.mark.django_db
def test_quoted_value_with_special_symbols(api_client, clear_cache):
    title = "'|(), '"
    create_book(title)

    response = filter_api(api_client, 'title={}'.format(title))
    assert_ok_response(response, 0)


@pytest.mark.django_db
def test_unquoted_value_with_special_symbols(api_client, clear_cache):
    title = '|(),"'
    with pytest.raises(RQLFilterParsingError):
        filter_api(api_client, 'title={}'.format(title))


@pytest.mark.django_db
def test_value_with_quotes_fail(api_client, clear_cache):
    title = '|\'(),"'
    with pytest.raises(RQLFilterParsingError):
        filter_api(api_client, 'title__exact={}'.format(title))


@pytest.mark.django_db
def test_empty_property(api_client, clear_cache):
    create_book()

    response = filter_api(api_client, '=')
    assert_ok_response(response, 1)


@pytest.mark.django_db
@pytest.mark.parametrize('prop', (
    'title',
    'title__exact',
))
def test_empty_value(api_client, clear_cache, prop):
    create_book()

    response = filter_api(api_client, '{}='.format(prop))
    assert_ok_response(response, 1)


@pytest.mark.django_db
def test_no_query(api_client, clear_cache):
    create_book()

    response = filter_api(api_client, '')
    assert_ok_response(response, 1)


@pytest.mark.django_db
def test_pagination(api_client, clear_cache):
    create_books(['G', 'G', 'H'])

    response = filter_api(api_client, 'limit=1&offset=1&title__exact=G')
    assert response['Content-Range'] == 'items 1-1/2'


@pytest.mark.django_db
def test_choice(api_client, clear_cache):
    books = [Book.objects.create(str_choice_field=choice) for choice, _ in Book.STR_CHOICES]

    response = filter_api(api_client, 'str_choice_field__in={},{}'.format('one', 'two'))
    assert_ok_response(response, 2)

    response = filter_api(api_client, 'str_choice_field__in=,{}'.format('one'))
    assert_ok_response(response, 1)
    assert response.data[0]['id'] == books[0].id



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from functools import partial

import pytest
from lark.exceptions import LarkError

from dj_rql.parser import RQLParser
from tests.test_parser.utils import LogicalTransformer
from dj_rql.constants import ComparisonOperators, LogicalOperators


def logical_transform(tpl, operator=None, exp1=None, exp2=None):
    return LogicalTransformer().transform(RQLParser.parse(tpl.format(
        operator=operator, exp1=exp1, exp2=exp2 or exp1,
    )))


def check_expression(logical_operator, expression_str, logical_result):
    grammar_key = LogicalOperators.get_grammar_key(logical_operator)
    assert grammar_key in logical_result
    for element in logical_result[grammar_key][0]:
        b = element in expression_str
        if element == ComparisonOperators.EQ:
            b = b or ('=' in expression_str)
        assert b

    if logical_operator != LogicalOperators.NOT:
        assert logical_result[grammar_key][0] == logical_result[grammar_key][1]


base_and_logical_transform = partial(logical_transform, '{operator}({exp1},{exp2})')
alias_and_logical_transform = partial(logical_transform, '{exp1}{operator}{exp2}')
base_or_logical_transform = partial(logical_transform, '{operator}({exp1},{exp2})')
alias_or_logical_transform = partial(logical_transform, '({exp1}{operator}{exp2})')
not_logical_transform = partial(logical_transform, '{operator}({exp1})', LogicalOperators.NOT)

ok_expressions = ['p=v', 'eq(p,v1)', 'p.p1=ge=25', 't__z="v v"']
fail_expressions = ['p=', 'and =ge=or', "'p=v'"]


@pytest.mark.parametrize('operator', ['&', ','])
@pytest.mark.parametrize('expression', ok_expressions)
def test_and_ok(operator, expression):
    base_result = base_and_logical_transform(LogicalOperators.AND, expression)
    check_expression(LogicalOperators.AND, expression, base_result)

    alias_result = alias_and_logical_transform(operator, expression)
    assert alias_result == base_result


@pytest.mark.parametrize('operator', ['|', ';'])
@pytest.mark.parametrize('expression', ok_expressions)
def test_or_ok(operator, expression):
    base_result = base_or_logical_transform(LogicalOperators.OR, expression)
    check_expression(LogicalOperators.OR, expression, base_result)

    alias_result = alias_or_logical_transform(operator, expression)
    assert alias_result == base_result


@pytest.mark.parametrize('expression', ok_expressions)
def test_not_ok(expression):
    base_result = not_logical_transform(expression)
    check_expression(LogicalOperators.NOT, expression, base_result)


@pytest.mark.parametrize('operator', ['&', ',', '|', ';'])
@pytest.mark.parametrize('expression', fail_expressions)
def test_and_or_expression_fail(operator, expression):
    with pytest.raises(LarkError):
        alias_or_logical_transform(operator, expression)


@pytest.mark.parametrize('expression', fail_expressions)
def test_not_fail(expression):
    with pytest.raises(LarkError):
        not_logical_transform(expression)


@pytest.mark.parametrize('query', [
    'p.p1=12&p', 'p=v|t=z', 'v=1,,f=g', 'and(or(v,k),b)', '(p=v|k=l)&n=m)',
])
def test_logical_fail(query):
    with pytest.raises(LarkError):
        logical_transform(query)


def test_logical_nesting():
    q = '((p1=v1|p1=ge=v2)&(p3=le=v4,gt(p5,v6));ne(p7,p8))'
    result = logical_transform(q)

    and_grammar_key = LogicalOperators.get_grammar_key(LogicalOperators.AND)
    or_grammar_key = LogicalOperators.get_grammar_key(LogicalOperators.OR)

    assert result == {
        or_grammar_key: [{
            and_grammar_key: [{
                or_grammar_key: [
                    (ComparisonOperators.EQ, 'p1', 'v1'),
                    (ComparisonOperators.GE, 'p1', 'v2'),
                ],
            }, {
                and_grammar_key: [
                    (ComparisonOperators.LE, 'p3', 'v4'),
                    (ComparisonOperators.GT, 'p5', 'v6'),
                ],
            }],
        }, (ComparisonOperators.NE, 'p7', 'p8')],
    }


def test_and_chain():
    q = 'ne(p1,v1)&p2=ge=and,or=v3'
    result = logical_transform(q)
    and_grammar_key = LogicalOperators.get_grammar_key(LogicalOperators.AND)
    assert result == {
        and_grammar_key: [
            (ComparisonOperators.NE, 'p1', 'v1'),
            {
                and_grammar_key: [
                    (ComparisonOperators.GE, 'p2', 'and'),
                    (ComparisonOperators.EQ, 'or', 'v3'),
                ],
            },
        ],
    }

    assert result == logical_transform('(ne(p1,v1)&(p2=ge=and,or=v3))')


def test_or_chain():
    q = '(ne(p1,v1)|(p2=ge=and;or=v3))'
    or_grammar_key = LogicalOperators.get_grammar_key(LogicalOperators.OR)
    assert logical_transform(q) == {
        or_grammar_key: [
            (ComparisonOperators.NE, 'p1', 'v1'),
            {
                or_grammar_key: [
                    (ComparisonOperators.GE, 'p2', 'and'),
                    (ComparisonOperators.EQ, 'or', 'v3'),
                ],
            },
        ],
    }



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from django.db.models import CharField, IntegerField, Value
from rest_framework import mixins
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from dj_rql.drf.backend import (
    RQLFilterBackend,
)
from dj_rql.drf.paginations import RQLContentRangeLimitOffsetPagination
from dj_rql.drf.compat import DjangoFiltersRQLFilterBackend
from tests.dj_rf.filters import BooksFilterClass, SelectBooksFilterClass
from tests.dj_rf.models import Book
from tests.dj_rf.serializers import SelectBookSerializer, BookSerializer


def apply_annotations(qs):
    return qs.annotate(
        anno_int=Value(1000, IntegerField()),
        anno_str=Value('text', CharField(max_length=10)),
    )


class BaseViewSet(mixins.ListModelMixin, GenericViewSet):
    queryset = apply_annotations(
        Book.objects.select_related('author').prefetch_related('pages').all(),
    )
    serializer_class = BookSerializer
    rql_filter_class = BooksFilterClass
    pagination_class = RQLContentRangeLimitOffsetPagination


class DRFViewSet(BaseViewSet):
    filter_backends = (RQLFilterBackend,)

    @action(detail=True)
    def act(self, *args):
        return Response()


class OpenAPIRetrieveSpecBackend(DjangoFiltersRQLFilterBackend):
    OPENAPI_RETRIEVE_SPECIFICATION = True


class DjangoFiltersViewSet(mixins.RetrieveModelMixin, mixins.UpdateModelMixin, BaseViewSet):
    filter_backends = (OpenAPIRetrieveSpecBackend,)


class SelectViewSet(mixins.RetrieveModelMixin, DRFViewSet):
    serializer_class = SelectBookSerializer
    rql_filter_class = SelectBooksFilterClass


class NoFilterClsViewSet(DRFViewSet):
    rql_filter_class = None



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest
from django.db.models import IntegerField, Value, Prefetch

from dj_rql.qs import (
    Annotation,
    Chain,
    DBOptimization,
    PrefetchRelated,
    NestedPrefetchRelated,
    NestedSelectRelated,
    SelectRelated,
    NPR,
    NSR,
    _NestedOptimizationMixin,
)
from tests.dj_rf.models import Book


default_qs = Book.objects.all()


def test_args():
    with pytest.raises(AssertionError) as e:
        DBOptimization()

    assert str(e.value) == 'At least one optimization must be specified.'


@pytest.mark.parametrize('relation', (['z'], ['z', 'x']))
def test_main_relation(relation):
    assert DBOptimization(*relation).main_relation


def test_extensions_no_extension():
    assert DBOptimization('z').extensions == {}


def test_extensions_annotations():
    assert DBOptimization('z', anno_1=1, anno_2=2).extensions == {
        'anno_1': 1, 'anno_2': 2,
    }


def test_base_apply():
    with pytest.raises(NotImplementedError):
        DBOptimization('z').apply(None)


@pytest.mark.django_db
def test_annotation_apply():
    Book.objects.create()

    anno = Annotation(abc=Value(1, IntegerField()))
    assert anno.apply(default_qs).first().abc == 1

    assert anno.rebuild() == anno


def test_sr_apply():
    qs = SelectRelated('author', 'author__publisher').apply(default_qs)
    assert qs.query.select_related == {'author': {'publisher': {}}}


def test_pr_apply():
    p_obj = Prefetch('pages')
    qs = PrefetchRelated('pages', p_obj).apply(default_qs)
    assert qs._prefetch_related_lookups == ('pages', p_obj)


def test_nested_deep_nesting():
    qs = NPR('authors').rebuild(NSR('publisher').rebuild(NSR('author'))).apply(default_qs)
    assert qs._prefetch_related_lookups == ('author__publisher__authors',)


def test_nested_chain_parent():
    qs = NSR('publisher').rebuild(Chain(Chain(NSR('author'), NPR('pages')))).apply(default_qs)
    assert qs.query.select_related == {'author': {'publisher': {}}}


def test_nested_apply_annotation_parent():
    qs = NestedPrefetchRelated('pages').rebuild(Annotation(**{'a': 1})).apply(default_qs)
    assert qs._prefetch_related_lookups == ('pages',)


def test_nested_prefetch_object_in_parent():
    with pytest.raises(AssertionError) as e:
        NestedPrefetchRelated('pages').rebuild(PrefetchRelated(Prefetch('pages')))

    assert str(e.value) == 'Only simple parent relations are supported.'


def test_npr_apply_no_parent():
    qs = NestedPrefetchRelated('pages').apply(default_qs)
    assert qs._prefetch_related_lookups == ('pages',)


@pytest.mark.parametrize('parent', (PrefetchRelated('author'), SelectRelated('author')))
def test_npr_apply_parent(parent):
    qs = NestedPrefetchRelated('publisher').rebuild(parent).apply(default_qs)
    assert qs._prefetch_related_lookups == ('author__publisher',)


def test_npr_apply_multi():
    qs = NPR(Prefetch('publisher'), 'publisher').rebuild(SelectRelated('author')).apply(default_qs)

    assert len(qs._prefetch_related_lookups) == 2
    assert qs._prefetch_related_lookups[1] == 'author__publisher'

    pr_obj = qs._prefetch_related_lookups[0]
    assert isinstance(pr_obj, Prefetch)
    assert pr_obj.prefetch_through == pr_obj.prefetch_to == 'author__publisher'


def test_nsr_apply_no_parent():
    qs = NestedSelectRelated('author', 'author__publisher').apply(default_qs)
    assert qs.query.select_related == {'author': {'publisher': {}}}


def test_nsr_apply_sr_parent():
    qs = NestedSelectRelated('publisher').rebuild(SelectRelated('author')).apply(default_qs)
    assert qs.query.select_related == {'author': {'publisher': {}}}


def test_nsr_apply_pr_parent():
    qs = NestedSelectRelated('publisher').rebuild(PrefetchRelated('author')).apply(default_qs)
    assert qs._prefetch_related_lookups == ('author__publisher',)


def test_chain_without_parent():
    qs = Chain(SelectRelated('author__publisher'), SelectRelated('author')).apply(default_qs)
    assert qs.query.select_related == {'author': {'publisher': {}}}


def test_chain_parent_sr():
    qs = Chain(NSR('publisher')).rebuild(Chain(SelectRelated('author'))).apply(default_qs)
    assert qs.query.select_related == {'author': {'publisher': {}}}


def test_chain_parent_pr():
    qs = Chain(NSR('publisher')).rebuild(Chain(PrefetchRelated('author'))).apply(default_qs)
    assert qs._prefetch_related_lookups == ('author__publisher',)


def test_chain_bad_relations():
    with pytest.raises(AssertionError) as e:
        Chain('abc')

    assert str(e.value) == 'Wrong Chain() optimization configuration.'


def test_nested_optimization_mixin_rebuild_nested():
    with pytest.raises(NotImplementedError):
        _NestedOptimizationMixin()._rebuild_nested(None)



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest
from django.db import connection
from django.test.utils import CaptureQueriesContext
from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK

from tests.dj_rf.models import Author, Book, Page, Publisher


@pytest.mark.django_db
def test_simple(api_client, clear_cache):
    Book.objects.create()
    response = api_client.get(reverse('select-list') + '?select(-id)')
    assert response.status_code == HTTP_200_OK
    assert 'id' not in response.data[0]


@pytest.mark.django_db
def test_complex(api_client, clear_cache):
    publisher = Publisher.objects.create(name='publisher')
    author = Author.objects.create(name='auth', publisher=publisher)
    book = Book.objects.create(author=author)
    Page.objects.create(book=book, number=1, content='text')

    with CaptureQueriesContext(connection) as context:
        response = api_client.get(reverse('select-list') + '?select(-author)')

    assert response.status_code == HTTP_200_OK
    assert 'author' not in response.data[0]
    assert 'author_ref' in response.data[0]
    assert context.captured_queries[1]['sql'].count('JOIN') == 1



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest
from django.test.utils import CaptureQueriesContext
from django.db import connection
from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK

from dj_rql.drf import RQLFilterBackend, FilterCache
from tests.dj_rf.models import Book


@pytest.mark.django_db
def test_quoting(api_client, clear_cache):
    books = [Book.objects.create(title='Title with ws'), Book.objects.create()]
    response = api_client.get(reverse('book-list') + '?title="Title%20with%20ws"')
    assert response.status_code == HTTP_200_OK
    assert response.data == [{'id': books[0].pk}]


@pytest.mark.django_db
def test_reflecting_star_in_like(api_client, clear_cache):
    books = [Book.objects.create(title=r'title\*'), Book.objects.create(title='title\\')]
    response = api_client.get(reverse('book-list') + r'?like(title,"title\\\*")')
    assert response.status_code == HTTP_200_OK
    assert response.data == [{'id': books[0].pk}]


@pytest.mark.django_db
def test_reflecting_20_and_minus(api_client, clear_cache):
    books = [Book.objects.create(title='A  B-c0')]
    response = api_client.get(reverse('book-list') + r'?like(title,*A%20 B-c0*)')
    assert response.status_code == HTTP_200_OK
    assert response.data == [{'id': books[0].pk}]


@pytest.mark.django_db
def test_list(api_client, clear_cache):
    books = [Book.objects.create() for _ in range(2)]
    response = api_client.get(reverse('book-list') + '?')
    assert response.status_code == HTTP_200_OK
    assert response.data == [{'id': books[0].pk}, {'id': books[1].pk}]


@pytest.mark.django_db
def test_list_empty(api_client, clear_cache):
    response = api_client.get(reverse('book-list'))
    assert response.status_code == HTTP_200_OK
    assert response.data == []


@pytest.mark.django_db
def test_list_filtering(api_client, clear_cache):
    books = [Book.objects.create() for _ in range(2)]
    query = 'id={}'.format(books[0].pk)
    response = api_client.get('{}?{}'.format(reverse('book-list'), query))
    assert response.status_code == HTTP_200_OK
    assert response.data == [{'id': books[0].pk}]


@pytest.mark.django_db
def test_list_pagination(api_client, clear_cache):
    books = [Book.objects.create() for _ in range(5)]
    query = 'limit=2,eq(offset,1)'
    response = api_client.get('{}?{}'.format(reverse('book-list'), query))
    assert response.status_code == HTTP_200_OK
    assert response.data == [{'id': books[1].pk}, {'id': books[2].pk}]
    assert response._headers['content-range'][1] == 'items 1-2/5'


@pytest.mark.django_db
def test_list_pagination_zero_limit(api_client, clear_cache):
    [Book.objects.create() for _ in range(5)]
    query = 'limit=0'
    response = api_client.get('{}?{}'.format(reverse('book-list'), query))
    assert response.status_code == HTTP_200_OK
    assert response.data == []
    assert response._headers['content-range'][1] == 'items 0-0/5'


def test_rql_filter_cls_is_not_set():
    class View:
        pass

    assert RQLFilterBackend().filter_queryset(None, 'str', View()) == 'str'


@pytest.mark.django_db
def test_cache(api_client, clear_cache):
    books = [
        Book.objects.create(title='F'),
        Book.objects.create(title='G'),
    ]

    assert FilterCache.CACHE == {}
    response = api_client.get('{}?{}'.format(reverse('book-list'), 'title=F'))
    assert response.data == [{'id': books[0].pk}]

    expected_cache_key = 'book.BooksFilterClass'
    assert expected_cache_key in FilterCache.CACHE
    cache_item_id = id(FilterCache.CACHE[expected_cache_key])

    response = api_client.get('{}?{}'.format(reverse('book-list'), 'title=G'))
    assert response.data == [{'id': books[1].pk}]

    assert expected_cache_key in FilterCache.CACHE
    assert id(FilterCache.CACHE[expected_cache_key]) == cache_item_id

    FilterCache.clear()
    assert FilterCache.CACHE == {}


@pytest.mark.django_db
def test_distinct_sequence(api_client, clear_cache):
    with CaptureQueriesContext(connection) as context:
        api_client.get('{}?{}'.format(reverse('book-list'), 'status=planning'))

        assert 'distinct' in context.captured_queries[1]['sql'].lower()

    with CaptureQueriesContext(connection) as context:
        api_client.get('{}?{}'.format(reverse('book-list'), 'title=abc'))

        assert 'distinct' not in context.captured_queries[1]['sql'].lower()



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#


def get_book_filter_cls_ordering_data():
    return {
        'author.email',
        'published.at',
        'd_id',
        'int_choice_field',
        'ordering_filter',
        'fsm',
        'anno_int',
        'anno_int_ref',
    }


def get_book_filter_cls_search_data():
    return {
        'title',
        'author.email',
        'author__email',
        'str_choice_field',
        'fsm',
        'anno_str',
        'anno_title_dynamic',
    }



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from tests.dj_rf.models import Book
from tests.dj_rf.view import apply_annotations

book_qs = apply_annotations(Book.objects.order_by('id'))


def create_books(count=2):
    Book.objects.bulk_create([Book() for _ in range(count)])
    books = list(book_qs)
    return books



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from dj_rql.transformer import BaseRQLTransformer


class ComparisonTransformer(BaseRQLTransformer):
    def comp(self, args):
        prop, operation, value = self._extract_comparison(args)
        return operation, prop, value


class LogicalTransformer(ComparisonTransformer):
    def logical(self, args):
        return {args[0].data: args[0].children}


class ListTransformer(BaseRQLTransformer):
    def listing(self, args):
        return (self._get_value(args[0]), self._get_value(args[1]),
                tuple(self._get_value(arg) for arg in args[2:]))


class SearchTransformer(BaseRQLTransformer):
    def searching(self, args):
        return tuple(self._get_value(args[index]) for index in range(3))


class OrderingTransformer(BaseRQLTransformer):
    def ordering(self, args):
        return tuple(args[1:])



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from unittest import TestCase

from rest_framework.pagination import PAGE_BREAK, PageLink
from rest_framework.request import Request
from rest_framework.test import APIRequestFactory

from dj_rql.exceptions import RQLFilterParsingError
from dj_rql.drf import RQLContentRangeLimitOffsetPagination

factory = APIRequestFactory()


class TestRQLLimitOffsetPagination(TestCase):
    """ An adopted copy of DRF pagination test case. """
    def setUp(self):
        class ExamplePagination(RQLContentRangeLimitOffsetPagination):
            default_limit = 10
            max_limit = 15

        self.pagination = ExamplePagination()
        self.queryset = range(1, 101)

    def paginate_queryset(self, request):
        return list(self.pagination.paginate_queryset(self.queryset, request))

    def get_paginated_content(self, queryset):
        response = self.pagination.get_paginated_response(queryset)
        return response.data

    def get_content_range_data(self, queryset):
        response = self.pagination.get_paginated_response(queryset)
        return response['Content-Range']

    def get_html_context(self):
        return self.pagination.get_html_context()

    def test_no_offset(self):
        request = Request(factory.get('/', {'limit': 5}))
        queryset = self.paginate_queryset(request)
        content = self.get_paginated_content(queryset)
        context = self.get_html_context()
        assert queryset == [1, 2, 3, 4, 5]
        assert content == [1, 2, 3, 4, 5]
        assert context == {
            'previous_url': None,
            'next_url': 'http://testserver/?limit=5&offset=5',
            'page_links': [
                PageLink('http://testserver/?limit=5', 1, True, False),
                PageLink('http://testserver/?limit=5&offset=5', 2, False, False),
                PageLink('http://testserver/?limit=5&offset=10', 3, False, False),
                PAGE_BREAK,
                PageLink('http://testserver/?limit=5&offset=95', 20, False, False),
            ],
        }
        assert self.pagination.display_page_controls

        content_range_data = self.get_content_range_data(queryset)
        assert content_range_data == 'items 0-4/100'

    def test_pagination_not_applied_if_limit_or_default_limit_not_set(self):
        class MockPagination(RQLContentRangeLimitOffsetPagination):
            default_limit = None

        request = Request(factory.get('/'))
        queryset = MockPagination().paginate_queryset(self.queryset, request)
        assert queryset is None

    def test_single_offset(self):
        """
        When the offset is not a multiple of the limit we get some edge cases:
        * The first page should still be offset zero.
        * We may end up displaying an extra page in the pagination control.
        """
        request = Request(factory.get('/', {'limit': 5, 'offset': 1}))
        queryset = self.paginate_queryset(request)
        content = self.get_paginated_content(queryset)
        context = self.get_html_context()
        assert queryset == [2, 3, 4, 5, 6]
        assert content == [2, 3, 4, 5, 6]
        assert context == {
            'previous_url': 'http://testserver/?limit=5',
            'next_url': 'http://testserver/?limit=5&offset=6',
            'page_links': [
                PageLink('http://testserver/?limit=5', 1, False, False),
                PageLink('http://testserver/?limit=5&offset=1', 2, True, False),
                PageLink('http://testserver/?limit=5&offset=6', 3, False, False),
                PAGE_BREAK,
                PageLink('http://testserver/?limit=5&offset=96', 21, False, False),
            ],
        }

        content_range_data = self.get_content_range_data(queryset)
        assert content_range_data == 'items 1-5/100'

    def test_first_offset(self):
        request = Request(factory.get('/', {'limit': 5, 'offset': 5}))
        queryset = self.paginate_queryset(request)
        content = self.get_paginated_content(queryset)
        context = self.get_html_context()
        assert queryset == [6, 7, 8, 9, 10]
        assert content == [6, 7, 8, 9, 10]
        assert context == {
            'previous_url': 'http://testserver/?limit=5',
            'next_url': 'http://testserver/?limit=5&offset=10',
            'page_links': [
                PageLink('http://testserver/?limit=5', 1, False, False),
                PageLink('http://testserver/?limit=5&offset=5', 2, True, False),
                PageLink('http://testserver/?limit=5&offset=10', 3, False, False),
                PAGE_BREAK,
                PageLink('http://testserver/?limit=5&offset=95', 20, False, False),
            ],
        }

        content_range_data = self.get_content_range_data(queryset)
        assert content_range_data == 'items 5-9/100'

    def test_middle_offset(self):
        request = Request(factory.get('/', {'limit': 5, 'offset': 10}))
        queryset = self.paginate_queryset(request)
        content = self.get_paginated_content(queryset)
        context = self.get_html_context()
        assert queryset == [11, 12, 13, 14, 15]
        assert content == [11, 12, 13, 14, 15]
        assert context == {
            'previous_url': 'http://testserver/?limit=5&offset=5',
            'next_url': 'http://testserver/?limit=5&offset=15',
            'page_links': [
                PageLink('http://testserver/?limit=5', 1, False, False),
                PageLink('http://testserver/?limit=5&offset=5', 2, False, False),
                PageLink('http://testserver/?limit=5&offset=10', 3, True, False),
                PageLink('http://testserver/?limit=5&offset=15', 4, False, False),
                PAGE_BREAK,
                PageLink('http://testserver/?limit=5&offset=95', 20, False, False),
            ],
        }

        content_range_data = self.get_content_range_data(queryset)
        assert content_range_data == 'items 10-14/100'

    def test_ending_offset(self):
        request = Request(factory.get('/', {'limit': 5, 'offset': 95}))
        queryset = self.paginate_queryset(request)
        content = self.get_paginated_content(queryset)
        context = self.get_html_context()
        assert queryset == [96, 97, 98, 99, 100]
        assert content == [96, 97, 98, 99, 100]
        assert context == {
            'previous_url': 'http://testserver/?limit=5&offset=90',
            'next_url': None,
            'page_links': [
                PageLink('http://testserver/?limit=5', 1, False, False),
                PAGE_BREAK,
                PageLink('http://testserver/?limit=5&offset=85', 18, False, False),
                PageLink('http://testserver/?limit=5&offset=90', 19, False, False),
                PageLink('http://testserver/?limit=5&offset=95', 20, True, False),
            ],
        }

        content_range_data = self.get_content_range_data(queryset)
        assert content_range_data == 'items 95-99/100'

    def test_erronous_offset(self):
        request = Request(factory.get('/', {'limit': 5, 'offset': 1000}))
        queryset = self.paginate_queryset(request)
        self.get_paginated_content(queryset)
        self.get_html_context()

    def test_invalid_offset(self):
        """
        An invalid offset query param should be treated as 0.
        """
        request = Request(factory.get('/', {'limit': 5, 'offset': 'invalid'}))
        queryset = self.paginate_queryset(request)
        assert queryset == [1, 2, 3, 4, 5]

    def test_invalid_limit(self):
        """
        An invalid limit query param should be ignored in favor of the default.
        """
        request = Request(factory.get('/', {'limit': 'invalid', 'offset': 0}))
        queryset = self.paginate_queryset(request)
        assert queryset == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    def test_zero_limit(self):
        """
        An zero limit query param should be ignored in favor of the default.
        """
        request = Request(factory.get('/', {'limit': 0, 'offset': 0}))
        queryset = self.paginate_queryset(request)
        assert queryset == []

    def test_max_limit(self):
        """
        The limit defaults to the max_limit when there is a max_limit and the
        requested limit is greater than the max_limit
        """
        offset = 50
        request = Request(factory.get('/', {'limit': '11235', 'offset': offset}))
        queryset = self.paginate_queryset(request)
        assert queryset == list(range(51, 66))

    def test_rql_operators(self):
        limit, offset = 1, 2
        request = Request(factory.get('/?search=0&limit=eq={},eq(offset,{})'.format(limit, offset)))
        queryset = self.paginate_queryset(request)
        assert queryset == [3]

    def assert_rql_parsing_error(self, query):
        request = Request(factory.get('/?{}'.format(query)))
        with self.assertRaises(RQLFilterParsingError) as e:
            self.paginate_queryset(request)
        assert e.exception.details['error'] == 'Limit and offset are set incorrectly.'

    def test_bad_rql_comparison_operator(self):
        self.assert_rql_parsing_error('limit=ge=1')

    def test_several_limit_parameters(self):
        self.assert_rql_parsing_error('limit=1,offset=1,eq(limit,2)')

    def test_several_offset_parameters(self):
        self.assert_rql_parsing_error('limit=1,offset=1,offset=eq=2')



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest
from lark.exceptions import LarkError

from dj_rql.constants import SearchOperators
from dj_rql.parser import RQLParser
from tests.test_parser.constants import FAIL_PROPS, FAIL_VALUES, OK_PROPS, OK_VALUES
from tests.test_parser.utils import SearchTransformer


search_operators = [SearchOperators.LIKE, SearchOperators.I_LIKE]

SEARCH_OKAY_VALUES = OK_VALUES[:]
SEARCH_OKAY_VALUES.extend(['*v', 'v*', '*v*', '*v*v', '*v*v*', '"*v v*"', 'v v'])


def search_transform(operator, prop, value):
    query = '{operator}({prop},{value})'.format(operator=operator, prop=prop, value=value)
    return SearchTransformer().transform(RQLParser.parse(query))


@pytest.mark.parametrize('operator', search_operators)
@pytest.mark.parametrize('prop', OK_PROPS)
@pytest.mark.parametrize('value', SEARCH_OKAY_VALUES)
def test_searching_ok(operator, prop, value):
    assert search_transform(operator, prop, value) == (operator, prop, value)


@pytest.mark.parametrize('operator', search_operators)
@pytest.mark.parametrize('prop', FAIL_PROPS)
def test_searching_property_fail(operator, prop):
    with pytest.raises(LarkError):
        search_transform(operator, prop, 'value')


@pytest.mark.parametrize('operator', search_operators)
@pytest.mark.parametrize('value', FAIL_VALUES)
def test_searching_value_fail(operator, value):
    with pytest.raises(LarkError):
        search_transform(operator, 'prop', value)



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest

from dj_rql.constants import FilterLookups, FilterTypes


def test_field_filter_type():
    custom_field = {}

    assert FilterTypes.field_filter_type(custom_field) == FilterTypes.STRING


@pytest.mark.parametrize('func', ('numeric', 'string', 'boolean'))
def test_filter_lookups_non_null(func):
    result = getattr(FilterLookups, func)()
    result.discard(FilterLookups.NULL)

    assert result == getattr(FilterLookups, func)(with_null=False)



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest
from lark.exceptions import LarkError

from dj_rql.constants import ListOperators
from dj_rql.parser import RQLParser
from tests.test_parser.constants import FAIL_PROPS, LIST_FAIL_VALUES, OK_PROPS, OK_VALUES
from tests.test_parser.utils import ListTransformer

REVERSED_OK_VALUES = reversed(OK_VALUES)
list_operators = [ListOperators.IN, ListOperators.OUT]


def list_transform(operator, prop, values):
    query = '{operator}({prop},({values}))'.format(
        operator=operator, prop=prop, values=','.join(values),
    )
    return ListTransformer().transform(RQLParser.parse(query))


@pytest.mark.parametrize('operator', list_operators)
@pytest.mark.parametrize('prop', OK_PROPS)
@pytest.mark.parametrize('v1,v2', zip(OK_VALUES, REVERSED_OK_VALUES))
def test_list_ok(operator, prop, v1, v2):
    for v in (v1, v2):
        assert list_transform(operator, prop, (v,)) == (operator, prop, (v,))
    assert list_transform(operator, prop, (v1, v2)) == (operator, prop, (v1, v2))


@pytest.mark.parametrize('operator', list_operators)
@pytest.mark.parametrize('prop', FAIL_PROPS)
def test_list_property_fail(operator, prop):
    with pytest.raises(LarkError):
        list_transform(operator, prop, ('value',))


@pytest.mark.parametrize('operator', list_operators)
@pytest.mark.parametrize('v1,v2', zip(OK_VALUES, LIST_FAIL_VALUES))
def test_list_value_fail(operator, v1, v2):
    with pytest.raises(LarkError):
        list_transform(operator, 'prop', (v1, v2))



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from functools import partial

import pytest

from lark.exceptions import LarkError

from dj_rql.constants import ComparisonOperators as CompOp
from dj_rql.parser import RQLParser
from tests.test_parser.constants import FAIL_PROPS, FAIL_VALUES, OK_PROPS, OK_VALUES
from tests.test_parser.utils import ComparisonTransformer


def cmp_transform(tpl, operator, prop, value):
    return ComparisonTransformer().transform(RQLParser.parse(tpl.format(
        operator=operator, prop=prop, value=value,
    )))


base_cmp_transform = partial(cmp_transform, '{operator}({prop},{value})')
alias_cmp_transform = partial(cmp_transform, '{prop}={operator}={value}')
eq_cmp_transform = partial(cmp_transform, '{prop}{operator}{value}', '=')


@pytest.mark.parametrize('operator', [CompOp.EQ, CompOp.GT, CompOp.LE])
@pytest.mark.parametrize('prop', OK_PROPS)
@pytest.mark.parametrize('value', OK_VALUES)
def test_comparison_ok(operator, prop, value):
    base_result = base_cmp_transform(operator, prop, value)
    assert base_result == (operator, prop, value)

    alias_result = alias_cmp_transform(operator, prop, value)
    assert alias_result == base_result


@pytest.mark.parametrize('operator', [CompOp.NE, CompOp.GE, CompOp.LT])
@pytest.mark.parametrize('prop', FAIL_PROPS)
@pytest.mark.parametrize('value', ['value'])
@pytest.mark.parametrize('func', [base_cmp_transform, alias_cmp_transform])
def test_comparison_property_fail(operator, prop, value, func):
    with pytest.raises(LarkError):
        func(operator, prop, value)


@pytest.mark.parametrize('operator', [CompOp.EQ, CompOp.GE, CompOp.LE])
@pytest.mark.parametrize('prop', ['prop'])
@pytest.mark.parametrize('value', FAIL_VALUES)
@pytest.mark.parametrize('func', [base_cmp_transform, alias_cmp_transform])
def test_comparison_value_fail(operator, prop, value, func):
    with pytest.raises(LarkError):
        func(operator, prop, value)


def test_comparison_order():
    assert base_cmp_transform(CompOp.EQ, CompOp.LE, CompOp.GT) != \
        alias_cmp_transform(CompOp.LE, CompOp.EQ, CompOp.GT)


@pytest.mark.parametrize('prop', OK_PROPS)
@pytest.mark.parametrize('value', OK_VALUES)
def test_comparison_eq_ok(prop, value):
    assert eq_cmp_transform(prop, value) == base_cmp_transform(CompOp.EQ, prop, value)


@pytest.mark.parametrize('prop', FAIL_PROPS)
@pytest.mark.parametrize('value', ['value'])
def test_comparison_eq_property_fail(prop, value):
    with pytest.raises(LarkError):
        eq_cmp_transform(prop, value)


@pytest.mark.parametrize('prop', ['prop'])
@pytest.mark.parametrize('value', FAIL_VALUES)
def test_comparison_eq_value_fail(prop, value):
    with pytest.raises(LarkError):
        eq_cmp_transform(prop, value)



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from __future__ import absolute_import, unicode_literals

from django.conf.urls import include, url
from rest_framework.routers import SimpleRouter

from tests.dj_rf.view import DjangoFiltersViewSet, DRFViewSet, SelectViewSet, NoFilterClsViewSet


router = SimpleRouter()
router.register(r'books', DRFViewSet, basename='book')
router.register(r'old_books', DjangoFiltersViewSet, basename='old_book')
router.register(r'select', SelectViewSet, basename='select')
router.register(r'nofiltercls', NoFilterClsViewSet, basename='nofiltercls')

urlpatterns = [
    url(r'^', include(router.urls)),
]



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest
from django.core.exceptions import FieldError
from django.db.models import CharField, IntegerField, Value

from dj_rql.exceptions import RQLFilterParsingError
from dj_rql.filter_cls import RQLFilterClass
from dj_rql.qs import AN, CH, PR, NPR, NSR, SR

from tests.dj_rf.models import Author, Book
from tests.test_filter_cls.utils import book_qs


class SelectFilterCls(RQLFilterClass):
    MODEL = Book
    SELECT = True
    FILTERS = ('id',)

    @property
    def heirarchy(self):
        return self.select_tree

    @property
    def exclusions(self):
        return self.default_exclusions


def test_init_no_select():
    class Cls(SelectFilterCls):
        SELECT = False
        FILTERS = (
            {
                'filter': 'hidden',
                'source': 'id',
                'hidden': True,
            },
        )

    instance = Cls(book_qs)
    assert not instance.heirarchy
    assert not instance.exclusions


def test_init_default_select():
    instance = SelectFilterCls(book_qs)
    assert set(instance.heirarchy.keys()) == {'id'}
    assert instance.heirarchy['id']['path'] == 'id'
    assert not instance.heirarchy['id']['qs']
    assert not instance.exclusions


def test_init_select():
    class Cls(SelectFilterCls):
        FILTERS = (
            'id',
            {
                'filter': 'ft',
                'source': 'id',
            },
            {
                'namespace': 'ns',
                'source': 'author',
                'filters': (
                    'id',
                    {
                        'filter': 'ft',
                        'source': 'id',
                    },
                    {
                        'namespace': 'ns',
                        'source': 'publisher',
                        'filters': ('id',),
                    },
                ),
            },
        )

    instance = Cls(book_qs)

    assert set(instance.heirarchy.keys()) == {'id', 'ft', 'ns'}
    assert not instance.heirarchy['id']['fields']
    assert not instance.heirarchy['id']['namespace']
    assert instance.heirarchy['id']['path'] == 'id'

    assert not instance.heirarchy['ft']['fields']
    assert not instance.heirarchy['ft']['namespace']
    assert instance.heirarchy['ft']['path'] == 'ft'

    top_ns = instance.heirarchy['ns']
    assert top_ns['namespace']
    assert top_ns['path'] == 'ns'
    assert set(top_ns['fields'].keys()) == {'id', 'ft', 'ns'}

    assert not top_ns['fields']['id']['fields']
    assert not top_ns['fields']['id']['namespace']
    assert top_ns['fields']['id']['path'] == 'ns.id'

    assert not top_ns['fields']['ft']['fields']
    assert not top_ns['fields']['ft']['namespace']
    assert top_ns['fields']['ft']['path'] == 'ns.ft'

    inner_ns = instance.heirarchy['ns']['fields']['ns']
    assert inner_ns['namespace']
    assert inner_ns['path'] == 'ns.ns'
    assert set(inner_ns['fields'].keys()) == {'id'}

    assert not inner_ns['fields']['id']['fields']
    assert not inner_ns['fields']['id']['namespace']
    assert inner_ns['fields']['id']['path'] == 'ns.ns.id'


def test_init_hidden():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'filter': 'vi',
                'source': 'id',
            },
            {
                'namespace': 'vn',
                'source': 'author',
                'filters': (
                    {
                        'filter': 'vi',
                        'source': 'id',
                    }, {
                        'filter': 'nvi',
                        'source': 'id',
                        'hidden': True,
                    },
                ),
            },
            {
                'namespace': 'nvn',
                'source': 'author',
                'hidden': True,
                'filters': (
                    {
                        'filter': 'vi',
                        'source': 'id',
                        'hidden': False,
                    },
                    {
                        'filter': 'nvi',
                        'source': 'id',
                        'hidden': True,
                    },
                    {
                        'namespace': 'vn',
                        'source': 'publisher',
                        'filters': ('id',)
                    },
                ),
            },
        )

    instance = Cls(book_qs)
    assert set(instance.heirarchy.keys()) == {'vi', 'vn', 'nvn'}
    assert not instance.heirarchy['vi']['hidden']
    assert not instance.heirarchy['vn']['hidden']
    assert instance.heirarchy['nvn']['hidden']

    assert set(instance.heirarchy['vn']['fields'].keys()) == {'vi', 'nvi'}
    assert not instance.heirarchy['vn']['fields']['vi']['hidden']
    assert instance.heirarchy['vn']['fields']['nvi']['hidden']

    assert set(instance.heirarchy['nvn']['fields'].keys()) == {'vi', 'nvi', 'vn'}
    assert not instance.heirarchy['nvn']['fields']['vi']['hidden']
    assert not instance.heirarchy['nvn']['fields']['vn']['hidden']
    assert instance.heirarchy['nvn']['fields']['nvi']['hidden']
    assert instance.heirarchy['vn']['fields']['nvi']['hidden']

    assert not instance.heirarchy['nvn']['fields']['vn']['fields']['id']['hidden']

    assert instance.exclusions == {'nvn', 'vn.nvi', 'nvn.nvi'}


def test_init_qs():
    class Cls(SelectFilterCls):
        FILTERS = (
            'id',
            {
                'filter': 'ft1',
                'source': 'id',
                'qs': SR('author'),
            },
            {
                'filter': 'ft2',
                'source': 'id',
                'qs': NSR('author'),
            },
            {
                'namespace': 'ns',
                'source': 'author',
                'qs': NPR('author', 'pages'),
                'filters': (
                    {
                        'filter': 'ft1',
                        'source': 'id',
                        'qs': SR('author__publisher'),
                    },
                    {
                        'filter': 'ft2',
                        'source': 'id',
                        'qs': NSR('publisher'),
                    },
                    {
                        'namespace': 'ns',
                        'source': 'publisher',
                        'qs': NPR('publisher'),
                        'filters': (
                            {
                                'filter': 'id',
                                'qs': AN(abc=Value(1, IntegerField())),
                            },
                        ),
                    },
                ),
            },
            {
                'namespace': 'pages',
                'qs': PR('pages'),
                'filters': (
                    {
                        'namespace': 'book',
                        'filters': (
                            {
                                'filter': 'author',
                                'source': 'author__id',
                                'qs': CH(NSR('author'), NSR('author__publisher')),
                            },
                        ),
                    },
                ),
            },
        )

    base_qs = Book.objects.all()
    instance = Cls(base_qs)

    assert instance.heirarchy['id']['qs'] is None

    qs = instance.heirarchy['ft1']['qs'].apply(base_qs)
    assert qs.query.select_related == {'author': {}}

    qs = instance.heirarchy['ft2']['qs'].apply(base_qs)
    assert qs.query.select_related == {'author': {}}

    top_ns = instance.heirarchy['ns']
    qs = top_ns['qs'].apply(base_qs)
    assert qs._prefetch_related_lookups == ('author', 'pages')

    qs = top_ns['fields']['ft1']['qs'].apply(base_qs)
    assert qs.query.select_related == {'author': {'publisher': {}}}

    qs = top_ns['fields']['ft2']['qs'].apply(base_qs)
    assert qs._prefetch_related_lookups == ('author__publisher',)

    inner_ns = top_ns['fields']['ns']
    qs = inner_ns['qs'].apply(base_qs)
    assert qs._prefetch_related_lookups == ('author__publisher',)

    qs = inner_ns['fields']['id']['qs'].apply(base_qs)
    assert set(qs.query.annotations.keys()) == {'abc'}

    pages = instance.heirarchy['pages']
    qs = pages['qs'].apply(base_qs)
    assert qs._prefetch_related_lookups == ('pages',)

    assert pages['fields']['book']['qs'] is None

    qs = pages['fields']['book']['fields']['author']['qs'].apply(base_qs)
    assert qs._prefetch_related_lookups == ('pages__author', 'pages__author__publisher')


class _Request:
    pass


def test_apply_rql_select_not_applied_non_select_cls():
    class Cls(SelectFilterCls):
        SELECT = False
        FILTERS = (
            {
                'filter': 'hidden',
                'source': 'id',
                'hidden': True,
            },
        )

    request = _Request()
    Cls(book_qs).apply_filters('select(+hidden)', request)
    assert hasattr(request, 'rql_ast')
    assert not hasattr(request, 'rql_select')


def test_apply_rql_select_applied_no_request():
    result = SelectFilterCls(book_qs).apply_filters('select(id)')
    assert result


def test_apply_rql_select_applied_no_query():
    request = _Request()

    SelectFilterCls(book_qs).apply_filters('', request)
    assert hasattr(request, 'rql_ast')
    assert request.rql_select == {'depth': 0, 'select': {}}


def test_default_exclusion_included():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'filter': 'ft1',
                'source': 'id',
                'hidden': True,
            },
            {
                'filter': 'ft2',
                'source': 'id',
            },
        )

    request = _Request()
    Cls(book_qs).apply_filters('select(-ft2)', request)
    assert request.rql_select['select'] == {'ft1': False, 'ft2': False}


def test_default_exclusion_overridden():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'filter': 'ft1',
                'source': 'id',
                'hidden': True,
            },
            {
                'filter': 'ft2',
                'source': 'id',
            },
        )

    request = _Request()
    Cls(book_qs).apply_filters('select(-ft2,ft1)', request)
    assert request.rql_select['select'] == {'ft1': True, 'ft2': False}


def test_signs_select():
    class Cls(SelectFilterCls):
        FILTERS = tuple(
            {
                'filter': 'ft{}'.format(i),
                'source': 'id',
            }
            for i in range(1, 5)
        )

    request = _Request()
    Cls(book_qs).apply_filters('select(ft1,+ft2,-ft3)', request)
    assert request.rql_select['select'] == {'ft1': True, 'ft2': True, 'ft3': False}


def test_bad_select_prop_top_level_include_select():
    with pytest.raises(RQLFilterParsingError) as e:
        SelectFilterCls(book_qs).apply_filters('select(abc)')

    assert e.value.details['error'] == 'Bad select filter: abc.'


def test_bad_select_prop_nested_include_select():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'namespace': 'ns',
                'source': 'author',
                'filters': ('id',),
            },
        )

    with pytest.raises(RQLFilterParsingError) as e:
        Cls(book_qs).apply_filters('select(+ns.abc)')

    assert e.value.details['error'] == 'Bad select filter: ns.abc.'


def test_bad_select_prop_top_level_exclude_select():
    with pytest.raises(RQLFilterParsingError) as e:
        SelectFilterCls(book_qs).apply_filters('select(-abc)')

    assert e.value.details['error'] == 'Bad select filter: -abc.'


def test_bad_select_prop_nested_exclude_select():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'namespace': 'ns',
                'source': 'author',
                'filters': ('id',),
            },
        )

    with pytest.raises(RQLFilterParsingError) as e:
        Cls(book_qs).apply_filters('select(-ns.abc)')

    assert e.value.details['error'] == 'Bad select filter: -ns.abc.'


def test_bad_select_prop_namespace_with_dot():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'namespace': 'ns',
                'source': 'author',
                'filters': ('id',),
            },
        )

    with pytest.raises(RQLFilterParsingError) as e:
        Cls(book_qs).apply_filters('select(+ns.)')

    assert e.value.details['error'] == 'Bad select filter: ns..'


def test_bad_select_prop_double_underscore_not_supported():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'namespace': 'ns',
                'source': 'author',
                'filters': ('id',),
            },
        )

    with pytest.raises(RQLFilterParsingError) as e:
        Cls(book_qs).apply_filters('select(+ns__id)')

    assert e.value.details['error'] == 'Bad select filter: ns__id.'


def test_bad_select_conf_included_then_excluded():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'namespace': 'ns',
                'source': 'author',
                'filters': ('id',),
            },
        )

    with pytest.raises(RQLFilterParsingError) as e:
        Cls(book_qs).apply_filters('select(ns.id,-ns)')

    assert e.value.details['error'] == 'Bad select filter: incompatible properties.'


def test_bad_select_conf_excluded_then_included():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'namespace': 'ns',
                'source': 'author',
                'filters': ('id',)
            },
        )

    with pytest.raises(RQLFilterParsingError) as e:
        Cls(book_qs).apply_filters('select(-ns,+ns.id)')

    assert e.value.details['error'] == 'Bad select filter: incompatible properties.'


def test_exclude_ok():
    request = _Request()

    SelectFilterCls(book_qs).apply_filters('select(-id)', request)
    assert request.rql_select == {'depth': 0, 'select': {'id': False}}


def test_select_complex():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'filter': 'ft1',
                'source': 'id',
            },
            {
                'filter': 'ft2',
                'source': 'id',
            },
            {
                'namespace': 'ns1',
                'source': 'author',
                'filters': (
                    {
                        'filter': 'ft1',
                        'source': 'id',
                        'hidden': True,
                    },
                    {
                        'filter': 'ft2',
                        'source': 'id',
                    },
                    {
                        'namespace': 'ns1',
                        'source': 'publisher',
                        'hidden': False,
                        'filters': ('id',)
                    },
                ),
            },
            {
                'namespace': 'ns2',
                'hidden': True,
                'source': 'author',
                'filters': (
                    {
                        'filter': 'ft1',
                        'source': 'id',
                        'hidden': True,
                    },
                    {
                        'filter': 'ft2',
                        'source': 'id',
                    },
                    {
                        'filter': 'ft3',
                        'source': 'id',
                    },
                    {
                        'namespace': 'ns1',
                        'source': 'publisher',
                        'filters': ('id',),
                    },
                    {
                        'namespace': 'ns2',
                        'source': 'publisher',
                        'filters': ('id',),
                    },
                ),
            },
        )

    request = _Request()
    Cls(book_qs).apply_filters('select(ns2.ns2.id,ft1,ns1,ns1.ft1,ns2.ft2)', request)
    assert request.rql_select['select'] == {
        'ft1': True,
        'ns1': True,
        'ns1.ft1': True,
        'ns2': True,
        'ns2.ft1': False,
        'ns2.ft2': True,
        'ns2.ft3': False,
        'ns2.ns2': True,
        'ns2.ns2.id': True,
        'ns2.ns1': False,
    }


def test_qs_optimization_full_tree():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'filter': 'ft1',
                'source': 'id',
                'qs': SR('author')
            },
            {
                'filter': 'ft2',
                'source': 'id',
                'qs': PR('pages'),
                'hidden': True,
            },
            {
                'namespace': 'ns',
                'source': 'author',
                'qs': NSR('author'),
                'filters': (
                    {
                        'filter': 'ft1',
                        'source': 'id',
                        'qs': NSR('publisher'),
                    },
                    {
                        'filter': 'ft2',
                        'source': 'id',
                        'qs': NPR('publisher'),
                    },
                ),
            },
        )

    _, qs = Cls(book_qs).apply_filters('select(-ns.ft2,ns)')
    assert qs.query.select_related == {'author': {'publisher': {}}}


def test_qs_optimization_custom_optimization():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'filter': 'ft1',
                'source': 'id',
            },
            {
                'filter': 'ft2',
                'source': 'id',
                'hidden': True,
            },
            {
                'namespace': 'ns',
                'source': 'author',
                'qs': NSR('author'),
                'filters': (
                    {
                        'filter': 'ft1',
                        'source': 'id',
                    },
                    {
                        'filter': 'ft2',
                        'source': 'id',
                    },
                ),
            },
        )

        def optimize_field(self, data):
            optimization_mapper = {
                'ft1': '_optimize_ft1',
                'ft2': '_optimize_ft2',
                'ns.ft1': '_optimize_ns_ft1',
                'ns.ft2': '_optimize_ns_ft2',
            }

            optimization_func = optimization_mapper.get(data.filter_path)
            if optimization_func:
                return getattr(self, optimization_func)(data)

        def _optimize_ft1(self, data):
            return data.queryset.select_related('author')

        def _optimize_ft2(self, data):
            return data.queryset.prefetch_related('pages')

        def _optimize_ns_ft1(self, data):
            return data.queryset.select_related('author__publisher')

        def _optimize_ns_ft2(self, data):
            return data.queryset.prefetch_related('author__publisher')

    _, qs = Cls(book_qs).apply_filters('select(-ns.ft2,ns)')
    assert qs.query.select_related == {'author': {'publisher': {}}}


def test_qs_optimization_django_error():
    class Cls(SelectFilterCls):
        FILTERS = (
            {
                'filter': 'ft1',
                'source': 'id',
                'qs': SR('invalid'),
            },
        )

    _, qs = Cls(book_qs).apply_filters('')

    with pytest.raises(FieldError):
        list(qs.all())


def test_qs_optimization_deep_various_nesting():
    class AuthorCls(SelectFilterCls):
        MODEL = Author
        FILTERS = (
            {
                'namespace': 'publisher',
                'qs': NSR('publisher'),
                'filters': (
                    {
                        'namespace': 'fk',
                        'source': 'fk1',
                        'qs': NSR('fk1'),
                        'filters': ('id',),
                    },
                    {
                        'namespace': 'fk2',
                        'source': 'fk2',
                        'qs': NSR('fk2'),
                        'filters': ('id',),
                    },
                    {
                        'namespace': 'authors',
                        'qs': NPR('authors'),
                        'filters': ('id',),
                    },
                ),
            },
            {
                'namespace': 'fk1',
                'qs': NSR('fk'),
                'filters': ('id',),
            },
            {
                'filter': 'id',
                'qs': NPR('books'),
            },
        )

    class BookCls(SelectFilterCls):
        MODEL = Book
        FILTERS = (
            {
                'namespace': 'author',
                'qs': NSR('author'),
                'filters': AuthorCls.FILTERS,
            },
        )

    _, qs = BookCls(book_qs).apply_filters('')
    assert qs.query.select_related == {'author': {'publisher': {'fk1': {}, 'fk2': {}}, 'fk': {}}}
    assert qs._prefetch_related_lookups == (
        'author__publisher__authors', 'author__books',
    )


@pytest.mark.django_db
def test_annotations():
    class AnnotationCls(SelectFilterCls):
        FILTERS = (
            {
                'namespace': 'ns',
                'source': 'author',
                'qs': AN(ns=Value('abc', CharField(max_length=128))),
                'filters': (
                    'id',
                    {
                        'filter': 'ft',
                        'source': 'id',
                        'qs': AN(ns_ft=Value('abc', CharField(max_length=128))),
                    },
                ),
            },
            {
                'filter': 'ft1',
                'source': 'id',
                'hidden': True,
                'qs': AN(ft1=Value(1, IntegerField())),
            },
            {
                'filter': 'ft2',
                'dynamic': True,
                'field': IntegerField(),
                'qs': AN(ft2=Value(1, IntegerField())),
            },
            {
                'filter': 'vns.ft',
                'source': 'vns_ft',
                'dynamic': True,
                'field': IntegerField(),
                'qs': AN(vns_ft=Value(1, IntegerField())),
            },
        )

    qs = Book.objects.all()

    instance = AnnotationCls(qs)
    assert set(instance.filters.keys()) == set(instance.annotations.keys()) - {'ns'}
    assert instance.annotations['ns'] == instance.annotations['ns.id']
    assert instance.annotations['ns'] != instance.annotations['ns.ft']

    _, qs = AnnotationCls(qs, instance=instance).apply_filters('')
    assert set(qs.query.annotations.keys()) == {'ns', 'ns_ft', 'ft2', 'vns_ft'}
    assert not list(qs)

    _, qs = AnnotationCls(qs, instance=instance).apply_filters('ft1=3&ft2=5&vns.ft=3')
    assert set(qs.query.annotations.keys()) == {'ns', 'ns_ft', 'ft1', 'ft2', 'vns_ft'}
    assert not list(qs)


def test_annotations_misconfiguration():
    class AnnotationCls(RQLFilterClass):
        MODEL = Book
        FILTERS = (
            {
                'filter': 'ft',
                'dynamic': True,
                'field': IntegerField(),
                'qs': AN(ft=Value(1, IntegerField())),
            },
        )

    with pytest.raises(FieldError):
        AnnotationCls(book_qs).apply_filters('ft=2')



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

import pytest
from django.core.exceptions import FieldDoesNotExist

from dj_rql.constants import FilterLookups as FL, RESERVED_FILTER_NAMES, RQL_NULL
from dj_rql.filter_cls import RQLFilterClass
from dj_rql.utils import assert_filter_cls
from tests.data import get_book_filter_cls_ordering_data, get_book_filter_cls_search_data
from tests.dj_rf.filters import BooksFilterClass
from tests.dj_rf.models import Author, Book

empty_qs = Author.objects.none()


def test_building_filters():
    non_null_string_lookups = FL.string()
    non_null_string_lookups.discard(FL.NULL)

    non_null_numeric_lookups = FL.numeric()
    non_null_numeric_lookups.discard(FL.NULL)

    expected_sub_dct = {
        'id': {'orm_route': 'id', 'lookups': FL.numeric()},
        'title': {
            'orm_route': 'title', 'lookups': FL.string(), 'null_values': {RQL_NULL, 'NULL_ID'},
        },
        'current_price': {
            'orm_route': 'current_price', 'lookups': FL.numeric(), 'null_values': {RQL_NULL},
        },
        'written': {'orm_route': 'written', 'lookups': FL.numeric()},
        'status': {'orm_route': 'status', 'lookups': non_null_string_lookups},
        'author__email': {'orm_route': 'author__email', 'lookups': FL.string()},
        'name': {'orm_route': 'author__name', 'lookups': FL.string()},
        'author.is_male': {'orm_route': 'author__is_male', 'lookups': FL.boolean()},
        'author.email': {'orm_route': 'author__email', 'lookups': FL.string()},
        'author.publisher.id': {
            'orm_route': 'author__publisher__id',
            'lookups': FL.numeric(),
        },
        'page.number': {'orm_route': 'pages__number', 'lookups': {FL.EQ, FL.NE}},
        'page.id': {'orm_route': 'pages__uuid', 'lookups': FL.string()},
        'published.at': {'orm_route': 'published_at', 'lookups': FL.numeric()},
        'rating.blog': {
            'orm_route': 'blog_rating', 'lookups': FL.numeric(), 'use_repr': True,
        },
        'rating.blog_int': {
            'orm_route': 'blog_rating', 'lookups': FL.numeric(), 'use_repr': False,
        },
        'amazon_rating': {
            'orm_route': 'amazon_rating', 'lookups': {FL.GE, FL.LT},
        },
        'url': {'orm_route': 'publishing_url', 'lookups': FL.string()},
        'd_id': [
            {'orm_route': 'id', 'lookups': FL.numeric()},
            {'orm_route': 'author__id', 'lookups': FL.numeric()},
        ],
        'custom_filter': {'custom': True, 'custom_data': [1], 'lookups': {FL.I_LIKE}},
        'int_choice_field': {
            'orm_route': 'int_choice_field', 'lookups': non_null_numeric_lookups,
        },
        'int_choice_field_repr': {
            'orm_route': 'int_choice_field', 'lookups': {FL.EQ, FL.NE}, 'use_repr': True,
        },
        'str_choice_field': {
            'orm_route': 'str_choice_field', 'lookups': non_null_string_lookups,
        },
        'str_choice_field_repr': {
            'orm_route': 'str_choice_field', 'lookups': {FL.EQ, FL.NE}, 'use_repr': True,
        },
        'has_list_lookup': {'custom': True, 'lookups': {FL.EQ, FL.IN, FL.OUT}},
        'no_list_lookup': {'custom': True, 'lookups': {FL.EQ}},
        't__in': {'orm_route': 'title', 'lookups': FL.string()},
        'github_stars': {'orm_route': 'github_stars', 'lookups': FL.numeric()},
        'ordering_filter': {'custom': True, 'ordering': True, 'lookups': {FL.EQ}},
        'fsm': {'orm_route': 'fsm_field', 'lookups': FL.string()},
        'anno_int': {'orm_route': 'anno_int', 'lookups': {FL.EQ}},
        'anno_int_ref': {'orm_route': 'anno_int', 'lookups': non_null_numeric_lookups},
        'anno_str': {'orm_route': 'anno_str', 'lookups': non_null_string_lookups},
        'anno_auto': {'orm_route': 'anno_auto', 'lookups': FL.numeric()},
        'anno_title_non_dynamic': {'orm_route': 'title', 'lookups': FL.string()},
        'anno_title_dynamic': {'orm_route': 'title', 'lookups': non_null_string_lookups},
        'author_publisher.id': {
            'orm_route': 'author__publisher__id',
            'lookups': FL.numeric(),
        },
    }

    assert_filter_cls(
        BooksFilterClass, expected_sub_dct,
        get_book_filter_cls_ordering_data(),
        get_book_filter_cls_search_data(),
    )


def test_bad_filter_configuration():
    with pytest.raises(AssertionError):
        assert_filter_cls(BooksFilterClass, {'prop': {}}, set(), set())


def test_model_is_not_set():
    with pytest.raises(AssertionError) as e:
        RQLFilterClass(empty_qs)
    assert str(e.value) == 'Model must be set for Filter Class.'


def test_wrong_extended_search_setup():
    class Cls(BooksFilterClass):
        EXTENDED_SEARCH_ORM_ROUTES = 'invalid'

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == 'Extended search ORM routes must be iterable.'


@pytest.mark.parametrize('filters', [None, {}, set()])
def test_fields_are_not_set(filters):
    class Cls(RQLFilterClass):
        MODEL = Author
        FILTERS = filters

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == 'List of filters must be set for Filter Class.'


def test_orm_path_misconfiguration():
    class Cls(RQLFilterClass):
        MODEL = Author
        FILTERS = ['base']

    with pytest.raises(FieldDoesNotExist):
        Cls(empty_qs)


def test_orm_field_type_is_unsupported():
    class Cls(RQLFilterClass):
        MODEL = Author
        FILTERS = ['publisher']

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == 'Unsupported field type: publisher.'


@pytest.mark.parametrize('filter_name', sorted(list(RESERVED_FILTER_NAMES)))
def test_reserved_filter_name_is_used(filter_name):
    class Cls(RQLFilterClass):
        MODEL = Author
        FILTERS = [{
            'filter': filter_name,
            'source': 'id',
        }]

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == "'{}' is a reserved filter name.".format(filter_name)


def test_bad_use_repr_and_ordering():
    class Cls(RQLFilterClass):
        MODEL = Book
        FILTERS = [{
            'filter': 'rating.blog',
            'source': 'blog_rating',
            'use_repr': True,
            'ordering': True,
        }]

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == "rating.blog: 'use_repr' and 'ordering' can't be used together."


def test_bad_use_repr_and_search():
    class Cls(RQLFilterClass):
        MODEL = Book
        FILTERS = [{
            'filter': 'str_choice_field',
            'use_repr': True,
            'search': True,
        }]

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == "str_choice_field: 'use_repr' and 'search' can't be used together."


@pytest.mark.parametrize('option', ('filter', 'dynamic', 'custom'))
def test_bad_option_in_namespace(option):
    class Cls(RQLFilterClass):
        MODEL = Book
        FILTERS = [{
            'namespace': 'title',
            option: True,
        }]

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == "title: '{}' is not supported by namespaces.".format(option)


def test_bad_item_structure():
    class Cls(RQLFilterClass):
        MODEL = Book
        FILTERS = [{
            'source': 'title',
        }]

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == "All extended filters must have set 'filter' set."


def test_bad_dynamic_in_namespace():
    class Cls(RQLFilterClass):
        MODEL = Book
        FILTERS = [{
            'namespace': 'author',
            'filters': [{
                'filter': 'a',
                'dynamic': True,
            }]
        }]

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == "author.a: dynamic filters are not supported in namespaces."


def test_dynamic_field_not_set():
    class Cls(RQLFilterClass):
        MODEL = Book
        FILTERS = [{
            'filter': 'title',
            'dynamic': True,
        }]

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == "title: dynamic filters must have 'field' set."


def test_bad_dynamic_set():
    class Cls(RQLFilterClass):
        MODEL = Book
        FILTERS = [{
            'filter': 'custom',
            'custom': True,
            'field': True,
            'lookups': {FL.EQ},
        }, {
            'filter': 'common',
            'field': True,
        }]

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == "common: common filters can't have 'field' set."


def test_bad_search():
    class Cls(RQLFilterClass):
        MODEL = Book
        FILTERS = [{
            'filter': 'id',
            'search': True,
        }]

    with pytest.raises(AssertionError) as e:
        Cls(empty_qs)
    assert str(e.value) == "id: 'search' can be applied only to text filters."


def test_get_field():
    class Cls(RQLFilterClass):
        MODEL = Book
        FILTERS = ['id']

        @classmethod
        def get_field(cls, *args):
            return cls._get_field(*args)

    assert Cls.get_field(Book, '') is None



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from django.db.models import IntegerField, CharField, AutoField, F

from dj_rql.drf.fields import SelectField
from dj_rql.filter_cls import RQLFilterClass
from dj_rql.constants import FilterLookups, RQL_NULL
from dj_rql.qs import AN, NSR, PR, SR
from tests.dj_rf.models import Book


AUTHOR_FILTERS = ['is_male', {
    'filter': 'email',
    'ordering': True,
    'search': True,
}, {
    'namespace': 'publisher',
    'filters': ['id'],
    'qs': NSR('publisher'),
}]


PAGE_FILTERS = [{
    'filter': 'number',
    'lookups': {FilterLookups.EQ, FilterLookups.NE},
}, {
    'filter': 'id',
    'source': 'uuid',
}]


class BooksFilterClass(RQLFilterClass):
    MODEL = Book
    FILTERS = ['id', {
        'filter': 'title',
        'null_values': {RQL_NULL, 'NULL_ID'},
        'search': True,
    }, 'current_price', 'written', {
        'filter': 'status',
        'distinct': True,
        'openapi': {
            'required': True,
        },
    }, {
        'filter': 'author__email',
        'search': True,
        'openapi': {
            'description': 'Author Email',
            'deprecated': True,
        },
    }, {
        'filter': 'name',
        'source': 'author__name',
        'distinct': True,
        'openapi': {
            'hidden': True,
        },
    }, {
        'namespace': 'author',
        'filters': AUTHOR_FILTERS,
        'qs': SR('author', 'author__publisher'),
    }, {
        'namespace': 'page',
        'source': 'pages',
        'filters': PAGE_FILTERS,
        'qs': PR('pages'),
    }, {
        'filter': 'published.at',
        'source': 'published_at',
        'ordering': True,
        'distinct': True,
    }, {
        'filter': 'rating.blog',
        'source': 'blog_rating',
        'use_repr': True,
    }, {
        'filter': 'rating.blog_int',
        'source': 'blog_rating',
        'use_repr': False,
    }, {
        'filter': 'amazon_rating',
        'lookups': {FilterLookups.GE, FilterLookups.LT},
        'null_values': {'random'},
    }, {
        'filter': 'url',
        'source': 'publishing_url',
        'openapi': {
            'type': 'string',
        },
    }, {
        'filter': 'd_id',
        'sources': ['id', 'author__id'],
        'ordering': True,
    }, {
        'filter': 'custom_filter',
        'distinct': True,
        'custom': True,
        'lookups': {FilterLookups.I_LIKE},

        'custom_data': [1],
    }, {
        'filter': 'int_choice_field',
        'ordering': True,
    }, {
        'filter': 'int_choice_field_repr',
        'source': 'int_choice_field',
        'use_repr': True,
        'lookups': {FilterLookups.EQ, FilterLookups.NE},
    }, {
        'filter': 'str_choice_field',
        'search': True,
    }, {
        'filter': 'str_choice_field_repr',
        'source': 'str_choice_field',
        'use_repr': True,
        'lookups': {FilterLookups.EQ, FilterLookups.NE},
    }, {
        'filter': 'has_list_lookup',
        'custom': True,
        'lookups': {FilterLookups.EQ, FilterLookups.NE, FilterLookups.IN, FilterLookups.OUT},
    }, {
        'filter': 'no_list_lookup',
        'custom': True,
        'lookups': {FilterLookups.EQ},
    }, {
        'filter': 't__in',
        'source': 'title',
        'openapi': {
            'type': 'custom',
            'format': 'custom',
        },
    }, 'github_stars', {
        'filter': 'ordering_filter',
        'custom': True,
        'ordering': True,
        'lookups': {FilterLookups.EQ},
    }, {
        'filter': 'fsm',
        'source': 'fsm_field',
        'ordering': True,
        'search': True,
    }, {
        'filter': 'anno_int',
        'dynamic': True,
        'field': IntegerField(),
        'lookups': {FilterLookups.EQ},
        'ordering': True,
    }, {
        'filter': 'anno_str',
        'dynamic': True,
        'field': CharField(),
        'search': True,
    }, {
        'filter': 'anno_int_ref',
        'dynamic': True,
        'field': IntegerField(),
        'source': 'anno_int',
        'ordering': True,
    }, {
        'filter': 'anno_auto',
        'dynamic': True,
        'field': AutoField(null=True),
        'qs': AN(anno_auto=F('id')),
    }, {
        'filter': 'anno_title_non_dynamic',
        'dynamic': False,
        'source': 'title',
    }, {
        'filter': 'anno_title_dynamic',
        'dynamic': True,
        'source': 'title',
        'field': CharField(),
        'search': True,
    }, {
        'namespace': 'author_publisher',
        'source': 'author__publisher',
        'filters': ['id'],
    }, {
        'filter': 'select_author',
        'dynamic': True,
        'field': SelectField(),
        'hidden': True,
        'qs': SR('author'),
    }]


class SelectBooksFilterClass(BooksFilterClass):
    SELECT = True



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from dj_rql.constants import (
    ComparisonOperators, ListOperators, LogicalOperators, SearchOperators,
    RQL_EMPTY, RQL_NULL,
)


OK_PROPS = (
    'p', 'p1', 'prop.p2', 'u_p', 'p1__p2__p3',
    ComparisonOperators.LT, LogicalOperators.AND, LogicalOperators.NOT, ListOperators.OUT,
    SearchOperators.LIKE,
)
OK_VALUES = [
    'value', 'PRD-000-000' "0", '""', "''", '-3.23', '+2', '"text , t lt"', 'val*',
    '2014-01-21T19:31:58+03:00', '2015-02-12', '"eq(1,2)"', 'email@example.com',
    r'\*', 'текст',
    ComparisonOperators.NE, LogicalOperators.OR, LogicalOperators.NOT, ListOperators.IN,
    SearchOperators.I_LIKE,
    RQL_NULL, RQL_EMPTY,
]
FAIL_PROPS = ['', '=', 't t', '"p"', '23', '1p', 'v*', '+v', 'текст']

LIST_FAIL_VALUES = ('', '"sdsd')
FAIL_VALUES = LIST_FAIL_VALUES + ('v1,v2',)



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

from uuid import uuid4

from django.db import models
from django_fsm import FSMField
from model_utils import Choices


class RandomFk(models.Model):
    pass


class Publisher(models.Model):
    name = models.CharField(max_length=20, null=True)

    fk1 = models.ForeignKey(RandomFk, on_delete=models.SET_NULL, null=True)
    fk2 = models.ForeignKey(RandomFk, on_delete=models.SET_NULL, null=True)


class Author(models.Model):
    name = models.CharField(max_length=20, null=True, blank=False)
    email = models.EmailField(null=True)

    is_male = models.NullBooleanField()

    publisher = models.ForeignKey(
        Publisher, related_name='authors', on_delete=models.SET_NULL, null=True,
    )

    fk1 = models.ForeignKey(RandomFk, on_delete=models.SET_NULL, null=True)


class Book(models.Model):
    LOW_RATING, HIGH_RATING = 0, 1
    BLOG_RATING_CHOICES = ((LOW_RATING, 'low'), (HIGH_RATING, 'high'))
    PLANNING, WRITING, PUBLISHED = 'planning', 'writing', 'publishing'
    STATUS_CHOICES = (PLANNING, WRITING, PUBLISHED)

    title = models.CharField(max_length=20, null=True, blank=True)
    status = models.CharField(max_length=20, choices=STATUS_CHOICES, default=PLANNING)
    publishing_url = models.URLField(null=True)

    blog_rating = models.BigIntegerField(null=True, choices=BLOG_RATING_CHOICES)
    github_stars = models.PositiveIntegerField(null=True)
    amazon_rating = models.FloatField(null=True)
    current_price = models.DecimalField(null=True, decimal_places=4)

    written = models.DateField(null=True)
    published_at = models.DateTimeField(null=True)

    author = models.ForeignKey(Author, related_name='books', on_delete=models.CASCADE, null=True)

    INT_CHOICES = Choices(
        (1, 'one', 'I'),
        (2, 'two', 'II'),
    )
    int_choice_field = models.IntegerField(choices=INT_CHOICES, default=INT_CHOICES.one)

    STR_CHOICES = Choices(
        ('one', 'I'),
        ('two', 'II'),
    )
    str_choice_field = models.CharField(max_length=5, choices=STR_CHOICES, default=STR_CHOICES.one)

    fsm_field = FSMField(default=STR_CHOICES.one, choices=STR_CHOICES, null=True)


class Page(models.Model):
    uuid = models.UUIDField(primary_key=True, default=uuid4, editable=False)
    content = models.TextField(null=True)

    number = models.IntegerField(null=True)
    book = models.ForeignKey(Book, related_name='pages', on_delete=models.CASCADE)
